# coding: utf-8
"""
The wntr.msx.elements module includes concrete implementations of the 
multi-species water quality model elements, including species, constants,
parameters, terms, and reactions.
"""

from __future__ import annotations

import logging
from typing import Any, Dict, Tuple, Union

from wntr.epanet.util import ENcomment, NoteType
from wntr.utils.disjoint_mapping import KeyExistsError

from .base import (
    ExpressionType,
    ReactionType,
    SpeciesType,
    VariableType,
    VariableBase,
    VariableValuesBase,
    ReactionBase,
    ReactionSystemBase,
)

logger = logging.getLogger(__name__)


class Species(VariableBase):
    """Biological or chemical species.

    Parameters
    ----------
    name
        Species name
    species_type
        Species type
    units
        Units of mass for this species, see :attr:`units` property.
    atol : float | None
        Absolute tolerance when solving this species' equations, by default
        None
    rtol : float | None
        Relative tolerance when solving this species' equations, by default
        None
    note
        Supplementary information regarding this variable, by default None
        (see :class:`~wntr.epanet.util.NoteType`)
    diffusivity
        Diffusivity of the species in water, by default None
    _vars
        Reaction system this species is a part of, by default None
    _vals
        Initial quality values for this species, by default None

    Raises
    ------
    KeyExistsError
        If the name has already been used
    TypeError
        If `atol` and `rtol` are not the same type
    ValueError
        If `atol` or `rtol` ≤ 0

    Notes
    -----
    EPANET-MSX requires that `atol` and `rtol` either both be omitted, or
    both be provided. In order to enforce this, the arguments passed for
    `atol` and `rtol` must both be None or both be positive values.
    """

    def __init__(
        self,
        name: str,
        species_type: Union[SpeciesType, str],
        units: str,
        atol: float = None,
        rtol: float = None,
        *,
        note: NoteType = None,
        diffusivity: float = None,
        _vars: ReactionSystemBase = None,
        _vals: VariableValuesBase = None,
    ) -> None:
        super().__init__(name, note=note)
        if _vars is not None and not isinstance(_vars, ReactionSystemBase):
            raise TypeError("Invalid type for _vars, {}".format(type(_vars)))
        if _vals is not None and not isinstance(_vals, InitialQuality):
            raise TypeError("Invalid type for _vals, {}".format(type(_vals)))
        if _vars is not None and name in _vars:
            raise KeyExistsError("This variable name is already taken")
        species_type = SpeciesType.get(species_type)
        if species_type is None:
            raise TypeError("species_type cannot be None")
        self._species_type: SpeciesType = species_type
        self._tolerances: Tuple[float, float] = None
        self.set_tolerances(atol, rtol)
        self.units: str = units
        """Units of mass for this species. For bulk species, concentration is
        this unit divided by liters, for wall species, concentration is this
        unit divided by the model's area-unit (see options).
        """
        self.diffusivity: float = diffusivity
        """Diffusivity of this species in water, if being used, by default None"""
        self._vars: ReactionSystemBase = _vars
        self._vals: InitialQuality = _vals

    def set_tolerances(self, atol: float, rtol: float):
        """Set the absolute and relative tolerance for the solvers.

        The user must set both values, or neither value (None). Values must be
        positive.

        Parameters
        ----------
        atol
            Absolute tolerance to use
        rtol
            Relative tolerance to use

        Raises
        ------
        TypeError
            If only one of `atol` or `rtol` is a float
        ValueError
            If `atol` or `rtol` ≤ 0
        """
        if (atol is None) ^ (rtol is None):
            raise TypeError("atol and rtol must both be float or both be None")
        if atol is None:
            self._tolerances = None
        elif atol <= 0 or rtol <= 0:
            raise ValueError("atol and rtol must both be positive, got atol={}, rtol={}".format(atol, rtol))
        else:
            self._tolerances = (atol, rtol)

    def get_tolerances(self) -> Union[Tuple[float, float], None]:
        """Get the custom solver tolerances for this species.

        Returns
        -------
        (atol, rtol) : (float, float) or None
            Absolute and relative tolerances, respectively, if they are set
        """
        return self._tolerances

    def clear_tolerances(self):
        """Set both tolerances to None, reverting to the global options value.
        """
        self._tolerances = None

    @property
    def atol(self) -> float:
        """Absolute tolerance. Must be set using :meth:`set_tolerances`"""
        if self._tolerances is not None:
            return self._tolerances[0]
        return None

    @property
    def rtol(self) -> float:
        """Relative tolerance. Must be set using :meth:`set_tolerances`"""
        if self._tolerances is not None:
            return self._tolerances[1]
        return None

    @property
    def var_type(self) -> VariableType:
        """Type of variable, :attr:`~wntr.msx.base.VariableType.SPECIES`."""
        return VariableType.SPECIES

    @property
    def species_type(self) -> SpeciesType:
        """Type of species, either :attr:`~wntr.msx.base.SpeciesType.BULK` 
        or :attr:`~wntr.msx.base.SpeciesType.WALL`"""
        return self._species_type

    @property
    def initial_quality(self) -> 'InitialQuality':
        """Initial quality values for the network"""
        if self._vals is not None:
            return self._vals
        else:
            raise TypeError("This species is not linked to a NetworkData obejct, please `relink` your model")

    @property
    def pipe_reaction(self) -> 'Reaction':
        """Pipe reaction definition"""
        if self._vars is not None:
            return self._vars.pipe_reactions[self.name]
        else:
            raise AttributeError("This species is not connected to a ReactionSystem")

    @property
    def tank_reaction(self) -> 'Reaction':
        """Tank reaction definition"""
        if self._vars is not None:
            return self._vars.tank_reactions[self.name]
        else:
            raise AttributeError("This species is not connected to a ReactionSystem")

    def to_dict(self) -> Dict[str, Any]:
        """Dictionary representation of the object

        The species dictionary has the following format, as described using a json schema.

        .. code:: json

            {
                "title": "Species",
                "type": "object",
                "properties": {
                    "name": {
                        "type": "string"
                    },
                    "species_type": {
                        "enum": ["bulk", "wall"]
                    },
                    "units": {
                        "type": "string"
                    },
                    "atol": {
                        "type": "number",
                        "exclusiveMinimum": 0
                    },
                    "rtol": {
                        "type": "number",
                        "exclusiveMinimum": 0
                    },
                    "note": {
                        "type": "string"
                    },
                    "diffusivity": {
                        "type": "number",
                        "minimum": 0
                    }
                },
                "required": ["name", "species_type", "units", "pipe_reaction", "tank_reaction"],
                "dependentRequired": {"atol": ["rtol"], "rtol":["atol"]}
            }

        """
        ret = dict(name=self.name, species_type=self.species_type.name.lower(), units=self.units, atol=self.atol, rtol=self.rtol)

        if self.diffusivity:
            ret["diffusivity"] = self.diffusivity

        if isinstance(self.note, ENcomment):
            ret["note"] = self.note.to_dict()
        elif isinstance(self.note, (str, dict, list)):
            ret["note"] = self.note

        return ret


class Constant(VariableBase):
    """Constant coefficient for use in expressions.

    Parameters
    ----------
    name
        Name of the variable.
    value
        Constant value.
    units
        Units for the variable, by default None
    note
        Supplementary information regarding this variable, by default None
    _vars
        Reaction system this constant is a part of, by default None
    """

    def __init__(self, name: str, value: float, *, units: str = None, note: Union[str, dict, ENcomment] = None, _vars: ReactionSystemBase = None) -> None:
        super().__init__(name, note=note)
        if _vars is not None and not isinstance(_vars, ReactionSystemBase):
            raise TypeError("Invalid type for _vars, {}".format(type(_vars)))
        if _vars is not None and name in _vars:
            raise KeyExistsError("This variable name is already taken")
        self.value: float = float(value)
        """Value of the constant"""
        self.units: str = units
        """Units of the constant"""
        self._vars: ReactionSystemBase = _vars

    def __call__(self, *, t=None) -> Any:
        return self.value

    @property
    def var_type(self) -> VariableType:
        """Type of variable, :attr:`~wntr.msx.base.VariableType.CONSTANT`."""
        return VariableType.CONSTANT

    def to_dict(self) -> Dict[str, Any]:
        """Dictionary representation of the object"""
        ret = dict(name=self.name, value=self.value)
        if self.units:
            ret["units"] = self.units
        if isinstance(self.note, ENcomment):
            ret["note"] = self.note.to_dict()
        elif isinstance(self.note, (str, dict, list)):
            ret["note"] = self.note
        return ret


class Parameter(VariableBase):
    """Parameterized variable for use in expressions.

    Parameters
    ----------
    name
        Name of this parameter.
    global_value
        Global value for the parameter if otherwise unspecified.
    units
        Units for this parameter, by default None
    note
        Supplementary information regarding this variable, by default None
    _vars
        Reaction system this parameter is a part of, by default None
    _vals
        Network-specific values for this parameter, by default None
    """
    def __init__(
        self, name: str, global_value: float, *, units: str = None, note: Union[str, dict, ENcomment] = None, _vars: ReactionSystemBase = None, _vals: VariableValuesBase = None
    ) -> None:
        super().__init__(name, note=note)
        if _vars is not None and not isinstance(_vars, ReactionSystemBase):
            raise TypeError("Invalid type for _vars, {}".format(type(_vars)))
        if _vals is not None and not isinstance(_vals, ParameterValues):
            raise TypeError("Invalid type for _vals, {}".format(type(_vals)))
        if _vars is not None and name in _vars:
            raise KeyExistsError("This variable name is already taken")
        self.global_value: float = float(global_value)
        self.units: str = units
        self._vars: ReactionSystemBase = _vars
        self._vals: ParameterValues = _vals

    def __call__(self, *, pipe: str = None, tank: str = None) -> float:
        """Get the value of the parameter for a given pipe or tank.

        If there is no specific, different value for the requested pipe
        or tank, then the global value will be returned. *This is true even*
        *if the pipe or tank does not exist in the network, so caution is*
        *advised*.

        Parameters
        ----------
        pipe
            Name of a pipe to get the parameter value for, by default None
        tank
            Name of a pipe to get the parameter value for, by default None

        Returns
        -------
        float
            Value at the specified pipe or tank, or the global value

        Raises
        ------
        TypeError
            If both pipe and tank are specified
        ValueError
            If there is no ParameterValues object defined for and linked to
            this parameter
        """
        if pipe is not None and tank is not None:
            raise TypeError("Both pipe and tank cannot be specified at the same time")
        elif self._vals is None and (pipe is not None or tank is not None):
            raise ValueError("No link provided to network-specific parameter values")
        if pipe:
            return self._vals.pipe_values.get(pipe, self.global_value)
        elif tank:
            return self._vals.tank_values.get(pipe, self.global_value)
        return self.global_value

    @property
    def var_type(self) -> VariableType:
        """Type of variable, :attr:`~wntr.msx.base.VariableType.PARAMETER`."""
        return VariableType.PARAMETER

    def to_dict(self) -> Dict[str, Any]:
        """Dictionary representation of the object"""
        ret = dict(name=self.name, global_value=self.global_value)
        if self.units:
            ret["units"] = self.units
        if isinstance(self.note, ENcomment):
            ret["note"] = self.note.to_dict()
        elif isinstance(self.note, (str, dict, list)):
            ret["note"] = self.note
        return ret


class Term(VariableBase):
    """Named expression (term) that can be used in expressions

    Parameters
    ----------
    name
        Variable name.
    expression
        Mathematical expression to be aliased
    note
        Supplementary information regarding this variable, by default None
    _vars
        Reaction system this species is a part of, by default None
    """
    def __init__(self, name: str, expression: str, *,
                 note: Union[str, dict, ENcomment] = None,
                 _vars: ReactionSystemBase = None) -> None:
        super().__init__(name, note=note)
        if _vars is not None and not isinstance(_vars, ReactionSystemBase):
            raise TypeError("Invalid type for _vars, {}".format(type(_vars)))
        if _vars is not None and name in _vars:
            raise KeyExistsError("This variable name is already taken")
        self.expression: str = expression
        """Expression that is aliased by this term"""
        self._vars: ReactionSystemBase = _vars

    @property
    def var_type(self) -> VariableType:
        """Type of variable, :attr:`~wntr.msx.base.VariableType.TERM`."""
        return VariableType.TERM

    def to_dict(self) -> Dict[str, Any]:
        """Dictionary representation of the object"""
        ret = dict(name=self.name, expression=self.expression)
        if isinstance(self.note, ENcomment):
            ret["note"] = self.note.to_dict()
        elif isinstance(self.note, (str, dict, list)):
            ret["note"] = self.note
        return ret


class ReservedName(VariableBase):
    """Reserved name that should not be used

    Parameters
    ----------
    name
        Reserved name.
    note
        Supplementary information regarding this variable, by default None

    Raises
    ------
    KeyExistsError
        If the name has already been registered
    """

    def __init__(self, name: str, *, note: Union[str, dict, ENcomment] = None) -> None:
        self.name: str = name
        self.note: Union[str, dict, ENcomment] = note

    @property
    def var_type(self) -> VariableType:
        """Type of variable, :attr:`~wntr.msx.base.VariableType.RESERVED`."""
        return VariableType.RESERVED

    # def to_dict(self) -> Dict[str, Any]:
    #     """Dictionary representation of the object"""
    #     return "{}({})".format(self.__class__.__name__, ", ".join(["name={}".format(repr(self.name)), "note={}".format(repr(self.note))]))


class HydraulicVariable(ReservedName):
    """Reserved name for hydraulic variables

    The user should not need to create any variables using this class, they
    are created automatically by the MsxModel object during initialization.

    Parameters
    ----------
    name
        Name of the variable (predefined by MSX)
    units
        Units for hydraulic variable, by default None
    note
        Supplementary information regarding this variable, by default None
    """

    def __init__(self, name: str, units: str = None, *, note: Union[str, dict, ENcomment] = None) -> None:
        super().__init__(name, note=note)
        self.units: str = units
        """Hydraulic variable's units"""


class MathFunction(ReservedName):
    """Reserved name for math functions

    Parameters
    ----------
    name
        Function name
    func
        Callable function
    note
        Supplementary information regarding this variable, by default None
    """

    def __init__(self, name: str, func: callable, *, note: Union[str, dict, ENcomment] = None) -> None:
        super().__init__(name, note=note)
        self.func: callable = func
        """A callable function or SymPy function"""

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        """Evaluate the function using any specified args and kwds."""
        return self.func(*args, **kwds)


class Reaction(ReactionBase):
    """Water quality reaction dynamics definition for a specific species.

    Parameters
    ----------
    species_name
        Species (object or name) this reaction is applicable to.
    reaction_type
        Reaction type (location), from {PIPE, TANK}
    expression_type
        Expression type (left-hand-side) of the equation, from 
        {RATE, EQUIL, FORMULA}
    expression
        Mathematical expression for the right-hand-side of the reaction
        equation.
    note
        Supplementary information regarding this variable, by default None
    _vars
        Reaction system this species is a part of, by default None
    """

    def __init__(
        self,
        species_name: str,
        reaction_type: ReactionType,
        expression_type: ExpressionType,
        expression: str,
        *,
        note: Union[str, dict, ENcomment] = None,
        _vars: ReactionSystemBase = None,
    ) -> None:
        super().__init__(species_name=species_name, note=note)
        if _vars is not None and not isinstance(_vars, ReactionSystemBase):
            raise TypeError("Invalid type for _vars, {}".format(type(_vars)))
        expression_type = ExpressionType.get(expression_type)
        reaction_type = ReactionType.get(reaction_type)
        if reaction_type is None:
            raise TypeError("Required argument reaction_type cannot be None")
        if expression_type is None:
            raise TypeError("Required argument expression_type cannot be None")
        self.__rxn_type: ReactionType = reaction_type
        self._expr_type: ExpressionType = expression_type
        if not expression:
            raise TypeError("expression cannot be None")
        self.expression: str = expression
        """Mathematical expression (right-hand-side)"""
        self._vars: ReactionSystemBase = _vars

    @property
    def expression_type(self) -> ExpressionType:
        """Expression type (left-hand-side), either 
        :attr:`~wntr.msx.base.ExpressionType.RATE`, 
        :attr:`~wntr.msx.base.ExpressionType.EQUIL`, or 
        :attr:`~wntr.msx.base.ExpressionType.FORMULA`"""
        return self._expr_type

    @property
    def reaction_type(self) -> ReactionType:
        """Reaction type (location), either 
        :attr:`~wntr.msx.base.ReactionType.PIPE` or  
        :attr:`~wntr.msx.base.ReactionType.TANK`"""
        return self.__rxn_type

    def to_dict(self) -> dict:
        """Dictionary representation of the object"""
        ret = dict(species_name=str(self.species_name), expression_type=self.expression_type.name.lower(), expression=self.expression)
        if isinstance(self.note, ENcomment):
            ret["note"] = self.note.to_dict()
        elif isinstance(self.note, (str, dict, list)):
            ret["note"] = self.note
        return ret


class InitialQuality(VariableValuesBase):
    """Initial quality values for a species in a specific network.

    Parameters
    ----------
    global_value
        Global initial quality value, by default 0.0
    node_values
        Any different initial quality values for specific nodes,
        by default None
    link_values
        Any different initial quality values for specific links,
        by default None
    """

    def __init__(self, global_value: float = 0.0, node_values: dict = None, link_values: dict = None):
        self.global_value = global_value
        """Global initial quality values for this species."""
        self._node_values = node_values if node_values is not None else dict()
        self._link_values = link_values if link_values is not None else dict()

    def __repr__(self) -> str:
        return self.__class__.__name__ + "(global_value={}, node_values=<{} entries>, link_values=<{} entries>)".format(
            self.global_value, len(self._node_values), len(self._link_values)
        )

    @property
    def var_type(self) -> VariableType:
        """Type of variable, :attr:`~wntr.msx.base.VariableType.SPECIES`,
        this object holds data for."""
        return VariableType.SPECIES

    @property
    def node_values(self) -> Dict[str, float]:
        """Mapping that overrides the global_value of the initial quality at
        specific nodes"""
        return self._node_values

    @property
    def link_values(self) -> Dict[str, float]:
        """Mapping that overrides the global_value of the initial quality in
        specific links"""
        return self._link_values

    def to_dict(self) -> Dict[str, Dict[str, float]]:
        """Dictionary representation of the object"""
        return dict(global_value=self.global_value, node_values=self._node_values.copy(), link_values=self._link_values.copy())


class ParameterValues(VariableValuesBase):
    """Pipe and tank specific values of a parameter for a specific network.

    Parameters
    ----------
    pipe_values
        Any different values for this parameter for specific pipes,
        by default None
    tank_values
        Any different values for this parameter for specific tanks,
        by default None
    """

    def __init__(self, *, pipe_values: dict = None, tank_values: dict = None) -> None:
        self._pipe_values = pipe_values if pipe_values is not None else dict()
        self._tank_values = tank_values if tank_values is not None else dict()

    def __repr__(self) -> str:
        return self.__class__.__name__ + "(pipe_values=<{} entries>, tank_values=<{} entries>)".format(len(self._pipe_values), len(self._tank_values))

    @property
    def var_type(self) -> VariableType:
        """Type of variable, :attr:`~wntr.msx.base.VariableType.PARAMETER`,
        this object holds data for."""
        return VariableType.PARAMETER

    @property
    def pipe_values(self) -> Dict[str, float]:
        """Mapping that overrides the global_value of a parameter for a
        specific pipe"""
        return self._pipe_values

    @property
    def tank_values(self) -> Dict[str, float]:
        """Mapping that overrides the global_value of a parameter for a
        specific tank"""
        return self._tank_values

    def to_dict(self) -> Dict[str, Dict[str, float]]:
        """Dictionary representation of the object"""
        return dict(pipe_values=self._pipe_values.copy(), tank_values=self._tank_values.copy())



import unittest
from os.path import abspath, dirname, join
import numpy as np
import pandas as pd
import wntr
from pandas.testing import assert_frame_equal, assert_series_equal

testdir = dirname(abspath(str(__file__)))
datadir = join(testdir, "networks_for_testing")
net3dir = join(testdir, "..", "..", "examples", "networks")


class TestPopulationMetrics(unittest.TestCase):
    def test_population_net3(self):
        inp_file = join(net3dir, "Net3.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        pop = wntr.metrics.population(wn)
        expected = 79000
        error = abs((pop.sum() - expected) / expected)
        self.assertLess(error, 0.01)  # 1% error

    def test_population_net6(self):
        inp_file = join(net3dir, "Net6.inp")
        wn = wntr.network.WaterNetworkModel(inp_file)
        pop = wntr.metrics.population(wn)
        expected = 152000
        error = abs((pop.sum() - expected) / expected)
        self.assertLess(error, 0.01)  # 1% error

    def test_population_impacted(self):  # tests query and population impacted

        pop = pd.Series([100, 200, 300, 400, 500], index=["J1", "J2", "J3", "J4", "J5"])

        # arg1 as a Series, arg2 as a scalar
        wsa = pd.Series([0.6, 0.7, 0.8, 0.9, 1], index=["J1", "J2", "J3", "J4", "J5"])
        pop_impacted = wntr.metrics.population_impacted(pop, wsa, np.less, 0.8)
        expected = pd.Series([100, 200, 0, 0, 0], index=["J1", "J2", "J3", "J4", "J5"])
        assert_series_equal(pop_impacted, expected, check_dtype=False)

        # arg1 as a Series, arg2 as a Series
        wsa = pd.Series([0.6, 0.7, 0.8, 0.9, 1], index=["J1", "J2", "J3", "J4", "J5"])
        wsa_threshold = pd.Series([1, 0, 1, 0, 1], index=["J1", "J2", "J3", "J4", "J5"])
        pop_impacted = wntr.metrics.population_impacted(
            pop, wsa, np.less_equal, wsa_threshold
        )
        expected = pd.Series(
            [100, 0, 300, 0, 500], index=["J1", "J2", "J3", "J4", "J5"]
        )
        assert_series_equal(pop_impacted, expected, check_dtype=False)

        # arg1 as a DataFrame, arg2 as a scalar
        wsa = pd.Series([0.6, 0.7, 0.8, 0.9, 1], index=["J1", "J2", "J3", "J4", "J5"])
        wsa = wsa.to_frame().transpose()
        wsa.loc[1, :] = [0, 1, 0, 1, 0]
        wsa.loc[2, :] = [1, 0, 1, 0, 1]
        pop_impacted = wntr.metrics.population_impacted(pop, wsa, np.less, 0.8)
        expected = pd.DataFrame(
            [[100, 200, 0, 0, 0], [100, 0, 300, 0, 500], [0, 200, 0, 400, 0]],
            columns=["J1", "J2", "J3", "J4", "J5"],
            index=[0, 1, 2],
        )
        assert_frame_equal(pop_impacted, expected, check_dtype=False)


if __name__ == "__main__":
    unittest.main()



"""
The wntr.gis.geospatial module contains functions to snap data and find 
intersects with polygons.
"""

import pandas as pd
import numpy as np
import matplotlib.pylab as plt
from scipy.spatial.distance import cdist
from scipy.spatial.distance import pdist, squareform
from scipy.cluster.hierarchy import linkage, fcluster, dendrogram

try:
    from shapely.geometry import MultiPoint, LineString, Point, shape
    has_shapely = True
except ModuleNotFoundError:
    has_shapely = False

try:
    import geopandas as gpd
    has_geopandas = True
except ModuleNotFoundError:
    gpd = None
    has_geopandas = False
    
try:
    import rasterio
    has_rasterio = True
except ModuleNotFoundError:
    rasterio = None
    has_rasterio = False


def snap(A, B, tolerance):  
    """
    Snap Points in A to Points or Lines in B

    For each Point geometry in A, the function returns snapped Point geometry 
    and associated element in B. Note the CRS of A must equal the CRS of B.
    
    Parameters
    ----------
    A : geopandas GeoDataFrame
        GeoDataFrame containing Point geometries.
    B : geopandas GeoDataFrame
        GeoDataFrame containing Point, LineString, or MultiLineString geometries.
    tolerance : float
        Maximum allowable distance (in the coordinate reference system units) 
        between Points in A and Points or Lines in B.  
    
    Returns
    -------
    GeoPandas GeoDataFrame
        Snapped points (index = A.index, columns = defined below)
        
        If B contains Points, columns include:
            - node: closest Point in B to Point in A
            - snap_distance: distance between Point in A and snapped point
            - geometry: GeoPandas Point object of the snapped point
        
        If B contains Lines or MultiLineString, columns include:
            - link: closest Line in B to Point in A
            - node: start or end node of Line in B that is closest to the snapped point (if B contains columns "start_node_name" and "end_node_name")
            - snap_distance: distance between Point A and snapped point
            - line_position: normalized distance of snapped point along Line in B from the start node (0.0) and end node (1.0)
            - geometry: GeoPandas Point object of the snapped point
    """   
    if not has_shapely or not has_geopandas:
        raise ModuleNotFoundError('shapley and geopandas are required')
        
    assert isinstance(A, gpd.GeoDataFrame)
    assert(A['geometry'].geom_type).isin(['Point']).all()
    assert isinstance(B, gpd.GeoDataFrame)
    assert (B['geometry'].geom_type).isin(['Point', 'LineString', 'MultiLineString']).all()
    assert A.crs == B.crs
    
    # Modify B to include "indexB" as a separate column
    B = B.reset_index(names='indexB')
    
    # Define the coordinate reference system, based on B
    crs = B.crs
    
    # Determine which Bs are closest to each A
    bbox = A.bounds + [-tolerance, -tolerance, tolerance, tolerance]       
    hits = bbox.apply(lambda row: list(B.loc[list(B.sindex.intersection(row))]['indexB']), axis=1)        
    closest = pd.DataFrame({
        # index of points table
        "point": np.repeat(hits.index, hits.apply(len)),
        # name of link
        "indexB": np.concatenate(hits.values)
        })
    
    # Merge the closest dataframe with the lines dataframe on the line names
    closest = pd.merge(closest, B, on="indexB")

    # Join back to the original points to get their geometry
    # rename the point geometry as "points"
    closest = closest.join(A.geometry.rename("points"), on="point")
    
    # Convert back to a GeoDataFrame, so we can do spatial ops
    closest = gpd.GeoDataFrame(closest, geometry="geometry", crs=crs)  
    
    # Calculate distance between the point and nearby links
    closest["snap_distance"] = closest.geometry.distance(gpd.GeoSeries(closest.points, crs=crs))
        
    # Collect only point/link pairs within snap distance radius
    # This is needed because B.sindex.intersection(row) above can return false positives
    closest = closest[closest['snap_distance'] <= tolerance]
    
    # Sort on ascending snap distance, so that closest goes to top
    closest = closest.sort_values(by=["snap_distance", "indexB"]) 
       
    # group by the index of the points and take the first, which is the closest line
    closest = closest.groupby("point").first()      
    
    # construct a GeoDataFrame of the closest elements of B
    closest = gpd.GeoDataFrame(closest, geometry="geometry", crs=crs)
    
    # Reset B index
    B.set_index('indexB', inplace=True)
    B.index.name = None
    
    # snap to points
    if B['geometry'].geom_type.isin(['Point']).all():
        snapped_points = closest.rename(columns={"indexB":"node"})
        snapped_points = snapped_points[["node", "snap_distance", "geometry"]]
        snapped_points.index.name = None      
        
    # snap to lines
    if B['geometry'].geom_type.isin(['LineString', 'MultiLineString']).all():
        closest = closest.rename(columns={"indexB":"link"})        
        # position of nearest point from start of the line
        pos = closest.geometry.project(gpd.GeoSeries(closest.points))        
        # get new point location geometry
        snapped_points = closest.geometry.interpolate(pos)
        snapped_points = gpd.GeoDataFrame(data=closest ,geometry=snapped_points, crs=crs)
        # determine whether the snapped point is closer to the start or end node
        snapped_points["line_position"] = closest.geometry.project(snapped_points, normalized=True)
        if ("start_node_name" in closest.columns) and ("end_node_name" in closest.columns):
            snapped_points.loc[snapped_points["line_position"]<0.5, "node"] = closest["start_node_name"]
            snapped_points.loc[snapped_points["line_position"]>=0.5, "node"] = closest["end_node_name"]
            snapped_points = snapped_points[["link", "node", "snap_distance", "line_position", "geometry"]]
        else:
            snapped_points = snapped_points[["link", "snap_distance", "line_position", "geometry"]]
        snapped_points.index.name = None
        
    return snapped_points

def _background(A, B):
    
    hull_geom = A.union_all().convex_hull
    hull_data = gpd.GeoDataFrame(pd.DataFrame([{'geometry': hull_geom}]), crs=A.crs)
    
    background_geom = hull_data.overlay(B, how='difference').union_all()
   
    background = gpd.GeoDataFrame(pd.DataFrame([{'geometry': background_geom}]), crs=A.crs)
    background.index = ['BACKGROUND']
    
    return background


def intersect(A, B, B_value=None, include_background=False, background_value=0):
    """
    Intersect Points, Lines or Polygons in A with Points, Lines, or Polygons in B.
    Return statistics on the intersection.
    
    The function returns information about the intersection for each geometry 
    in A. Each geometry in B is assigned a value based on a column of data in 
    that GeoDataFrame.  Note the CRS of A must equal the CRS of B.
    
    Parameters
    ----------
    A : geopandas GeoDataFrame
        GeoDataFrame containing Point, LineString, or Polygon geometries
    B : geopandas GeoDataFrame
        GeoDataFrame containing  Point, LineString, or Polygon geometries
    B_value : str or None (optional)
        Column name in B used to assign a value to each geometry.
        Default is None.
    include_background : bool (optional) 
        Include background, defined as space covered by A that is not covered by B 
        (overlay difference between A and B). The background geometry is added
        to B and is given the name 'BACKGROUND'. Default is False.
    background_value : int or float (optional)
        The value given to background space. This value is used in the intersection 
        statistics if a B_value column name is provided. Default is 0.
      
    Returns
    -------
    intersects : DataFrame
        Intersection statistics (index = A.index, columns = defined below)
        
        Columns include:
            - n: number of intersecting B geometries
            - intersections: list of intersecting B indices
            
        If B_value is given:
            - values: list of intersecting B values
            - sum: sum of the intersecting B values
            - min: minimum of the intersecting B values
            - max: maximum of the intersecting B values
            - mean: mean of the intersecting B values
            
        If A contains Lines and B contains Polygons:
            - weighted_mean: weighted mean of intersecting B values

    """
    if not has_shapely or not has_geopandas:
        raise ModuleNotFoundError('shapley and geopandas are required')
        
    assert isinstance(A, gpd.GeoDataFrame)
    assert (A['geometry'].geom_type).isin(['Point', 'LineString', 
                                           'MultiLineString', 'Polygon', 
                                           'MultiPolygon']).all()
    assert isinstance(B, gpd.GeoDataFrame)
    assert (B['geometry'].geom_type).isin(['Point', 'LineString', 
                                           'MultiLineString', 'Polygon', 
                                           'MultiPolygon']).all()
    if isinstance(B_value, str):
        assert B_value in B.columns
    assert isinstance(include_background, bool)
    assert isinstance(background_value, (int, float))
    assert A.crs == B.crs, "A and B must have the same crs."
    
    if include_background:
        background = _background(A, B)
        if B_value is not None:
            background[B_value] = background_value
        B = pd.concat([B, background])
    
    B_original_index_name = B.index.name
    B.index.name = None

    intersects = gpd.sjoin(A, B, predicate='intersects')
    intersects.index.name = '_tmp_index_name' # set a temp index name for grouping
    
    # Sort values by index and intersecting object
    intersects['sort_order'] = 1 # make sure 'BACKGROUND' is listed first
    intersects.loc[intersects['index_right'] == 'BACKGROUND', 'sort_order'] = 0
    intersects.sort_values(['_tmp_index_name', 'sort_order', 'index_right'], inplace=True)
    
    n = intersects.groupby('_tmp_index_name')['geometry'].count()
    B_indices = intersects.groupby('_tmp_index_name')['index_right'].apply(list)
    stats = pd.DataFrame(index=A.index.copy(), data={'intersections': B_indices,
                                              'n': n,})
    stats['n'] = stats['n'].fillna(0)
    stats['n'] = stats['n'].apply(int)
    stats.loc[stats['intersections'].isnull(), 'intersections'] = stats.loc[stats['intersections'].isnull(), 'intersections'] .apply(lambda x: [])
    
    if B_value is not None:
        stats['values'] = intersects.groupby('_tmp_index_name')[B_value].apply(list)
        stats['sum'] = intersects.groupby('_tmp_index_name')[B_value].sum()
        stats['min'] = intersects.groupby('_tmp_index_name')[B_value].min()
        stats['max'] = intersects.groupby('_tmp_index_name')[B_value].max()
        stats['mean'] = intersects.groupby('_tmp_index_name')[B_value].mean()
        
        stats = stats.reindex(['intersections', 'values', 'n', 'sum', 'min', 'max', 'mean'], axis=1)
        stats.loc[stats['values'].isnull(), 'values'] = stats.loc[stats['values'].isnull(), 'values'] .apply(lambda x: [])
        
    weighted_mean = False
    if (A['geometry'].geom_type).isin(['LineString', 'MultiLineString']).all():
        if (B['geometry'].geom_type).isin(['Polygon', 'MultiPolygon']).all():
            weighted_mean = True
            
    if weighted_mean and B_value is not None:
        stats['weighted_mean'] = 0.0
        A_length = A.length
        covered_length = pd.Series(0.0, index = A.index)
        
        for i in B.index:
            B_geom = gpd.GeoDataFrame(B.loc[[i],:], crs=B.crs)
            val = float(B_geom.iloc[0][B_value])
            A_subset = A.loc[stats['intersections'].apply(lambda x: i in x),:]
            #print(i, lines_subset)
            A_clip = gpd.clip(A_subset, B_geom) 
            A_clip_length = A_clip.length
            A_clip_index = A_clip.index
            
            if A_clip_length.shape[0] > 0:
                fraction_length = A_clip_length/A_length[A_clip_index]
                covered_length[A_clip_index] = covered_length[A_clip_index] + fraction_length
                weighed_val = fraction_length*val
                stats.loc[A_clip_index, 'weighted_mean'] = stats.loc[A_clip_index, 'weighted_mean'] + weighed_val
        
        # Normalize weighted mean by covered length (can be over 1 if polygons overlap)
        # Can be less than 1 if there are gaps (when background is not used)
        stats['weighted_mean'] = stats['weighted_mean']/covered_length
        
        # Covered_length is NaN if length A is 0, set weighted mean to mean
        stats.loc[covered_length.isna(), 'weighted_mean'] = stats.loc[covered_length.isna(), 'mean']
        
        # No intersection, set weighted mean to NaN
        stats.loc[stats['n']==0, 'weighted_mean'] = np.nan
        
    stats.index.name = None
    
    # Restore B's index
    B.index.name = B_original_index_name
    
    return stats


def sample_raster(A, filepath, bands=1):
    """Sample a raster (e.g., GeoTIFF file) using Points in GeoDataFrame A. 
    
    This function can take either a filepath to a raster or a virtual raster 
    (VRT), which combines multiple raster tiles into a single object. The 
    function opens the raster(s) and samples it at the coordinates of the point 
    geometries in A. This function assigns nan to values that match the 
    raster's `nodata` attribute. These sampled values are returned as a Series 
    which has an index matching A.

    Parameters
    ----------
    A : GeoDataFrame
        GeoDataFrame containing Point geometries
    filepath : str
        Path to raster or alternatively a virtual raster (VRT)
    bands : int or list[int] (optional, default = 1)
        Index or indices of the bands to sample (using 1-based indexing)

    Returns
    -------
    Series
        Pandas Series containing the sampled values for each geometry in gdf
    """
    # further functionality could include other geometries (Line, Polygon),
    # and use of multiprocessing to speed up querying.
    if not has_rasterio:
        raise ModuleNotFoundError('rasterio is required')
    
    assert (A['geometry'].geom_type == "Point").all()
    assert isinstance(filepath, str)
    assert isinstance(bands, (int, list))
    
    with rasterio.open(filepath) as raster:
        xys = zip(A.geometry.x, A.geometry.y)
        
        values = np.array(
            tuple(raster.sample(xys, bands)), dtype=float # force to float to allow for conversion of nodata to nan
        ).squeeze()
        
    values[values == raster.nodata] = np.nan
    values = pd.Series(values, index=A.index)
    
    return values


def connect_lines(lines, threshold, plot=False):
    """
    Connect lines by identifying start and end nodes that are within a 
    threshold distance

    Parameters
    ----------
    lines : gpd.GeoDataFrame
        GeoDataFrame with LineString geometry
    tolerance : float
        Maximum distance between line endpoints, used to define connecting 
        Point geometry
    plot : bool
        Boolean indicating if a plot is created for the dendogram, and original and connected lines
        
    Returns
    -------
    Tuple[line GeoDataFrame, node GeoDataFrame]
        * line GeoDataFrame contains LineString geometry, start_node_name, and 
          end_node_name, along with original columns in lines
        * node GeoDataFrame contains connecting Point geometry and node names
    """
    if not has_shapely or not has_geopandas:
        raise ModuleNotFoundError('shapley and geopandas are required')
    
    original_lines = lines
    lines = lines.copy()

    # Create start and end node name and Point geometry for each line
    nodes = []
    geometry = []
    lines['start_node_name'] = None #create new columns
    lines['end_node_name'] = None
    j = 0
    for i, line in lines.iterrows(): # loop through every line
        try:
            geom = line.geometry.geoms[0]
        except:
            geom = line.geometry
        
        geometry.append(Point([geom.coords[0][0], geom.coords[0][1]])) 
        nodes.append({'Line': i, 'Node': j}) #node list holds the line ID
        #unique number for each start node and end node for each line 
        lines.loc[i,'start_node_name'] = j
        j = j+1
            
        geometry.append(Point([geom.coords[-1][0], geom.coords[-1][1]]))
        nodes.append({'Line': i, 'Node': j})
        lines.loc[i,'end_node_name'] = j
        j = j+1
            
    nodes = gpd.GeoDataFrame(nodes, geometry=geometry)
    nodes.set_crs(lines.crs, inplace=True)
    
    # Create a distance matrix
    points = nodes['geometry']
    coords = points.apply(lambda geom: (geom.x, geom.y)).tolist()
    condensed = pdist(coords)
    D = squareform(condensed)
    D = pd.DataFrame(D, index=nodes.index, columns=nodes.index)
    
    # Several options exist for linkage and fcluster below 
    # currently using ward/euclidean/distance
    
    # Compute a linkage model
    Z = linkage(condensed, method='ward', metric='euclidean', optimal_ordering=True)
    
    # Form clusters
    #clusters = fcluster(Z, threshold, criterion='inconsistent', depth=2)
    clusters = fcluster(Z, threshold, criterion='distance')
    #clusters = fcluster(Z, threshold, criterion='monocrit')
    #clusters = fcluster(Z, threshold, criterion='maxclust')
    #clusters = fcluster(Z, threshold, criterion='maxclust_monocrit')
    clusters = pd.Series(clusters, index=nodes.index)
    nodes['supernode'] = clusters.copy()
    
    # Update lines GeoDataFrame with start and end supernode names
    map_node_to_supernode = nodes['supernode']
    lines['start_node_name'] = map_node_to_supernode[lines['start_node_name']].values
    lines['end_node_name'] = map_node_to_supernode[lines['end_node_name']].values
    
    # Remove lines with the same start and end node name
    lines = lines.loc[~(lines['start_node_name'] == lines['end_node_name']),:]
    
    # Create a nodes GeoDataFrame with centroid of each supernode
    supernode_name = []
    supernode_geom = []
    for name, group in nodes.groupby('supernode'):
        supernode_name.append(name)
        supernode_geom.append(group.dissolve().centroid[0])
    nodes = gpd.GeoDataFrame({'Node': supernode_name}, geometry=supernode_geom)
    nodes.set_index('Node', inplace=True)
    nodes.index.name = None
    nodes.crs = lines.crs
    
    # Convert names to string
    nodes.index = nodes.index.astype(str)
    lines['start_node_name'] = lines['start_node_name'].astype(str)
    lines['end_node_name'] = lines['end_node_name'].astype(str)
    
    # Add start and end node Points to LineStrings
    start_coords = nodes.loc[lines['start_node_name']]
    end_coords = nodes.loc[lines['end_node_name']]
    for i, (line_name, row) in enumerate(lines.iterrows()):
        line = row['geometry']
        l_coords = list(line.coords)
        start_point = nodes.loc[row['start_node_name'], 'geometry']
        end_point = nodes.loc[row['end_node_name'], 'geometry']
        if start_point.coords[0] != l_coords[0]:
            l_coords.insert(0, start_point.coords[0])
        if end_point.coords[0] != l_coords[-1]:
            l_coords.append(end_point.coords[0])
        lines.loc[line_name, 'geometry'] = LineString(l_coords)
    
    if plot:
  
        plt.figure(figsize=(10, 5))
        dendro = dendrogram(Z)

        #unique_clusters = np.unique(clusters)
        #colors = plt.cm.get_cmap('tab10', len(unique_clusters))  # Use a colormap with enough colors
        #color_dict = {cluster: colors(i) for i, cluster in enumerate(unique_clusters)}
        #dendro = dendrogram(Z, color_threshold=threshold, link_color_func=lambda k: color_dict[k])

        plt.title('Hierarchical Cluster Dendrogram')
        plt.xlabel('Data Point Indexes')
        plt.ylabel('Distance')
        
        plt.figure()
        ax = plt.gca()
        ax = original_lines.plot(ax=ax, color='r', label='Disconnected lines')
        ax = lines.plot(ax=ax, color='k', linewidth=6, alpha=0.35, label='Connected lines')
        ax = nodes.plot(ax=ax, color='k', label='Connected nodes')
        ax.legend()
        bounds = ax.axis('equal')
        plt.tight_layout()
        plt.axis('off')
    
    return lines, nodes



"""WNTR AML base classes."""

import sys
import scipy
from .evaluator import Evaluator
from .expr import Var, Param, native_numeric_types, Float, ConditionalExpression
from collections import OrderedDict
from wntr.utils.ordered_set import OrderedSet
from collections.abc import MutableMapping


class Constraint(object):
    __slots__ = ('_expr', 'name', '_c_obj')

    def __init__(self, expr):
        """

        Parameters
        ----------
        expr: wntr.sim.aml.expr.ExpressionBase
        """
        self._expr = expr
        self.name = None
        self._c_obj = None

    @property
    def expr(self):
        return self._expr

    @property
    def index(self):
        if self._c_obj is None:
            return None
        else:
            return self._c_obj.index

    def evaluate(self):
        return self.expr.evaluate()

    def reverse_ad(self):
        return self.expr.reverse_ad()


class Model(object):
    """
    A class for creating algebraic models.
    """
    def __init__(self):
        self._evaluator = Evaluator()
        self._refcounts = OrderedDict()
        self._con_ccon_map = OrderedDict()
        self._var_cvar_map = OrderedDict()
        self._param_cparam_map = OrderedDict()
        self._float_cfloat_map = OrderedDict()
        self._vars_referenced_by_con = OrderedDict()
        self._params_referenced_by_con = OrderedDict()
        self._floats_referenced_by_con = OrderedDict()

    def __setattr__(self, name, val):
        """
        Override built in __setattr__ so that params, vars, etc. get put in the appropriate dictionary

        Parameters
        ----------
        name: str
            name of the attribute
        val: object
            value of the attribute

        Returns
        -------
        None
        """
        if isinstance(val, (Var, Param, Constraint, _NodeDict)):
            if hasattr(self, name):
                raise ValueError('Model already has a {0} named {1}. If you want to replace the {0}, please remove the existing one first.'.format(type(val), name))

        if type(val) == Constraint:
            val.name = name
            self._register_constraint(val)
        elif type(val) == ConstraintDict:
            val.name = name
            val._model = self
            for k, v in val.items():
                self._register_constraint(v)
        elif type(val) in {Var, Param, VarDict, ParamDict}:
            val.name = name

        # The __setattr__ of the parent class should always be called so that the attribute actually gets set.
        super(Model, self).__setattr__(name, val)

    def __delattr__(self, name):
        """
        Override built in __delattr__ so that params, vars, etc. get removed from the appropriate dictionary

        Parameters
        ----------
        name: str
            name of the attribute

        Returns
        -------
        None
        """
        # The __delattr__ of the parent class should always be called so that the attribute actually gets removed.
        val = getattr(self, name)
        if type(val) == Constraint:
            self._remove_constraint(val)
            val.name = 'None'
        elif type(val) == ConstraintDict():
            val.name = 'None'
            val._model = None
            for k, v in val.items():
                self._remove_constraint(v)
        elif type(val) in {Var, Param, VarDict, ParamDict}:
            val.name = 'None'

        super(Model, self).__delattr__(name)

    def _increment_var(self, var):
        if var not in self._var_cvar_map:
            cvar = self._evaluator.add_var(var.value)
            var._c_obj = cvar
            self._var_cvar_map[var] = cvar
            self._refcounts[var] = 1
        else:
            self._refcounts[var] += 1
            cvar = self._var_cvar_map[var]
        return cvar

    def _increment_param(self, param):
        if param not in self._param_cparam_map:
            cparam = self._evaluator.add_param(param.value)
            param._c_obj = cparam
            self._param_cparam_map[param] = cparam
            self._refcounts[param] = 1
        else:
            self._refcounts[param] += 1
            cparam = self._param_cparam_map[param]
        return cparam

    def _increment_float(self, f):
        if f not in self._float_cfloat_map:
            cfloat = self._evaluator.add_float(f.value)
            f._c_obj = cfloat
            self._float_cfloat_map[f] = cfloat
            self._refcounts[f] = 1
        else:
            self._refcounts[f] += 1
            cfloat = self._var_cvar_map[f]
        return cfloat

    def _decrement_var(self, var):
        self._refcounts[var] -= 1
        if self._refcounts[var] == 0:
            cvar = self._var_cvar_map[var]
            var._c_obj = None
            var._value = cvar.value
            del self._refcounts[var]
            del self._var_cvar_map[var]
            self._evaluator.remove_var(cvar)

    def _decrement_param(self, p):
        self._refcounts[p] -= 1
        if self._refcounts[p] == 0:
            cparam = self._param_cparam_map[p]
            p._c_obj = None
            p._value = cparam.value
            del self._refcounts[p]
            del self._param_cparam_map[p]
            self._evaluator.remove_param(cparam)

    def _decrement_float(self, f):
        self._refcounts[f] -= 1
        if self._refcounts[f] == 0:
            cfloat = self._float_cfloat_map[f]
            f._c_obj = None
            del self._refcounts[f]
            del self._float_cfloat_map[f]
            self._evaluator.remove_float(cfloat)

    def _register_conditional_constraint(self, con):
        ccon = self._evaluator.add_if_else_constraint()
        con._c_obj = ccon
        self._con_ccon_map[con] = ccon
        leaf_ndx_map = OrderedDict()
        referenced_vars = OrderedSet()
        referenced_params = OrderedSet()
        referenced_floats = OrderedSet()
        ndx = 0
        derivs = list()
        for expr in con.expr._conditions:
            referenced_vars.update(expr.get_vars())
            referenced_params.update(expr.get_params())
            referenced_floats.update(expr.get_floats())
        for expr in con.expr._exprs:
            referenced_vars.update(expr.get_vars())
            referenced_params.update(expr.get_params())
            referenced_floats.update(expr.get_floats())
        for expr in con.expr._exprs:
            _deriv = expr.reverse_sd()
            derivs.append(_deriv)
            for v in referenced_vars:
                if v not in _deriv:
                    _deriv[v] = Float(0)
                elif type(_deriv[v]) in native_numeric_types:
                    _deriv[v] = Float(_deriv[v])
                referenced_floats.update(_deriv[v].get_floats())

        for v in referenced_vars:
            leaf_ndx_map[v] = ndx
            ndx += 1
            cvar = self._increment_var(v)
            ccon.add_leaf(cvar)
        for v in referenced_params:
            leaf_ndx_map[v] = ndx
            ndx += 1
            cvar = self._increment_param(v)
            ccon.add_leaf(cvar)
        for v in referenced_floats:
            leaf_ndx_map[v] = ndx
            ndx += 1
            cvar = self._increment_float(v)
            ccon.add_leaf(cvar)

        for i in range(len(con.expr._conditions)):
            condition_rpn = con.expr._conditions[i].get_rpn(leaf_ndx_map)
            for term in condition_rpn:
                ccon.add_condition_rpn_term(term)
            fn_rpn = con.expr._exprs[i].get_rpn(leaf_ndx_map)
            for term in fn_rpn:
                ccon.add_fn_rpn_term(term)
            for v in referenced_vars:
                cvar = v._c_obj
                jac = derivs[i][v]
                jac_rpn = jac.get_rpn(leaf_ndx_map)
                for term in jac_rpn:
                    ccon.add_jac_rpn_term(cvar, term)
            ccon.end_condition()

        self._vars_referenced_by_con[con] = referenced_vars
        self._params_referenced_by_con[con] = referenced_params
        self._floats_referenced_by_con[con] = referenced_floats

    def _register_constraint(self, con):
        if type(con.expr) == ConditionalExpression:
            self._register_conditional_constraint(con)
            return None
        ccon = self._evaluator.add_constraint()
        con._c_obj = ccon
        self._con_ccon_map[con] = ccon
        leaf_ndx_map = OrderedDict()
        referenced_vars = OrderedSet()
        referenced_params = OrderedSet()
        referenced_floats = OrderedSet()
        ndx = 0
        for v in con.expr.get_vars():
            leaf_ndx_map[v] = ndx
            ndx += 1
            cvar = self._increment_var(v)
            ccon.add_leaf(cvar)
            referenced_vars.add(v)
        for p in con.expr.get_params():
            leaf_ndx_map[p] = ndx
            ndx += 1
            cparam = self._increment_param(p)
            ccon.add_leaf(cparam)
            referenced_params.add(p)
        for f in con.expr.get_floats():
            leaf_ndx_map[f] = ndx
            ndx += 1
            cfloat = self._increment_float(f)
            ccon.add_leaf(cfloat)
            referenced_floats.add(f)
        fn_rpn = con.expr.get_rpn(leaf_ndx_map)
        for term in fn_rpn:
            ccon.add_fn_rpn_term(term)
        jac = con.expr.reverse_sd()
        for v in con.expr.get_vars():
            jac_v = jac[v]
            if type(jac_v) in native_numeric_types:
                jac_v = Float(jac_v)
            for f in jac_v.get_floats():
                if f not in leaf_ndx_map:
                    leaf_ndx_map[f] = ndx
                    ndx += 1
                    cfloat = self._increment_float(f)
                    ccon.add_leaf(cfloat)
                    referenced_floats.add(f)
            jac_rpn = jac_v.get_rpn(leaf_ndx_map)
            cvar = self._var_cvar_map[v]
            for term in jac_rpn:
                ccon.add_jac_rpn_term(cvar, term)
        self._vars_referenced_by_con[con] = referenced_vars
        self._params_referenced_by_con[con] = referenced_params
        self._floats_referenced_by_con[con] = referenced_floats

    def _remove_conditional_constraint(self, con):
        self._evaluator.remove_if_else_constraint(self._con_ccon_map[con])
        del self._con_ccon_map[con]
        for v in self._vars_referenced_by_con[con]:
            self._decrement_var(v)
        for p in self._params_referenced_by_con[con]:
            self._decrement_param(p)
        for f in self._floats_referenced_by_con[con]:
            self._decrement_float(f)
        del self._vars_referenced_by_con[con]
        del self._params_referenced_by_con[con]
        del self._floats_referenced_by_con[con]

    def _remove_constraint(self, con):
        if type(con.expr) == ConditionalExpression:
            self._remove_conditional_constraint(con)
            return None
        self._evaluator.remove_constraint(self._con_ccon_map[con])
        del self._con_ccon_map[con]
        for v in self._vars_referenced_by_con[con]:
            self._decrement_var(v)
        for p in self._params_referenced_by_con[con]:
            self._decrement_param(p)
        for f in self._floats_referenced_by_con[con]:
            self._decrement_float(f)
        del self._vars_referenced_by_con[con]
        del self._params_referenced_by_con[con]
        del self._floats_referenced_by_con[con]

    def evaluate_residuals(self, x=None):
        if x is not None:
            self._evaluator.load_var_values_from_x(x)
        r = self._evaluator.evaluate(len(self._con_ccon_map))
        return r

    def evaluate_jacobian(self, x=None):
        n_vars = len(self._var_cvar_map)
        n_cons = len(self._con_ccon_map)
        if n_vars != n_cons:
            raise ValueError('The number of constraints and variables must be equal.')
        if x is not None:
            self._evaluator.load_var_values_from_x(x)
        jac_values, col_ndx, row_nnz = self._evaluator.evaluate_csr_jacobian(self._evaluator.nnz,
                                                                             self._evaluator.nnz,
                                                                             len(self._con_ccon_map) + 1)
        result = scipy.sparse.csr_matrix((jac_values, col_ndx, row_nnz), shape=(n_cons, n_vars))
        return result

    def get_x(self):
        return self._evaluator.get_x(len(self._var_cvar_map))

    def load_var_values_from_x(self, x):
        self._evaluator.load_var_values_from_x(x)

    def __str__(self):
        tmp = 'cons:\n'
        for con in self._con_ccon_map.keys():
            tmp += str(con.name)
            tmp += ':   '
            tmp += str(con.expr)
            tmp += '\n'
        tmp += '\n'
        tmp += 'vars:\n'
        for var in self._var_cvar_map:
            tmp += str(var.name)
            tmp += ':   '
            tmp += str(var)
            tmp += '\n'
        return tmp

    def set_structure(self):
        """
        This method essentially just orders all of the variables and constraints so that
        the constraint residuals and the jacobian can be evaluated efficiently. This method
        must be called before get_x, load_var_values_from_x, evaluate_residuals, or evaluate_jacobian
        can be called. If any changes are made to the model (e.g., variables/constraints are
        added/removed), then this method needs called again. Avoid calling this method too often
        if you are concerned about efficiency.
        """
        self._evaluator.set_structure()

    def cons(self):
        for i in self._con_ccon_map:
            yield i

    def vars(self):
        for i in self._var_cvar_map:
            yield i


class _NodeDict(MutableMapping):
    def __init__(self, mapping=None):
        self._name = 'None'
        self._data = OrderedDict()

        if mapping is not None:
            self.update(mapping)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, val):
        self._name = val
        for k, v in self.items():
            v.name = self.name + '[' + str(k) + ']'

    def __delitem__(self, key):
        self._data[key].name = None
        del self._data[key]

    def __getitem__(self, key):
        return self._data[key]

    def __iter__(self):
        return self._data.__iter__()

    def __len__(self):
        return len(self._data)

    def __repr__(self):
        return self._data.__repr__()

    def __setitem__(self, key, val):
        val.name = self.name + '[' + str(key) + ']'
        self._data[key] = val

    def __str__(self):
        return self.__repr__()


class ParamDict(_NodeDict):
    pass


class VarDict(_NodeDict):
    pass


class ConstraintDict(_NodeDict):
    """
    Dictionary of constraints; primarily handles registering the constraints with the model and naming
    """
    def __init__(self, mapping=None):
        self._model = None
        super(ConstraintDict, self).__init__(mapping)

    def __delitem__(self, key):
        val = self[key]
        if self._model is not None:
            self._model._remove_constraint(val)
        val.name = 'None'
        del self._data[key]

    def __setitem__(self, key, val):
        if key in self:
            raise ValueError('ConstraintDict already has a Constraint named {0}. If you want to replace the Constraint, please remove the existing one first.'.format(key))
        val.name = self.name + '[' + str(key) + ']'
        if self._model is not None:
            self._model._register_constraint(val)
        self._data[key] = val



# coding: utf-8
"""
The wntr.msx.model module includes methods to build a multi-species water
quality model.
"""

from __future__ import annotations

import logging
import warnings
from typing import Dict, Generator, List, Union
from wntr.epanet.util import NoteType

from wntr.utils.disjoint_mapping import KeyExistsError

from .base import (
    EXPR_FUNCTIONS,
    HYDRAULIC_VARIABLES,
    QualityModelBase,
    ReactionBase,
    NetworkDataBase,
    ReactionSystemBase,
    ExpressionType,
    ReactionType,
    SpeciesType,
    VariableType,
)
from .elements import Constant, HydraulicVariable, InitialQuality, MathFunction, Parameter, ParameterValues, Reaction, Species, Term
from .options import MsxSolverOptions

logger = logging.getLogger(__name__)

MsxVariable = Union[Constant, HydraulicVariable, MathFunction, Parameter, Species, Term]
"""A class that is a valid MSX variable class"""


class MsxReactionSystem(ReactionSystemBase):
    """Registry for all the variables registered in the multi-species reactions
    model.

    This object can be used like a mapping.
    """

    def __init__(self) -> None:
        super().__init__()
        self._vars.add_disjoint_group("reserved")
        self._species = self._vars.add_disjoint_group("species")
        self._const = self._vars.add_disjoint_group("constant")
        self._param = self._vars.add_disjoint_group("parameter")
        self._term = self._vars.add_disjoint_group("term")
        self._rxns = dict(pipe=dict(), tank=dict())
        self._pipes = self._rxns["pipe"]
        self._tanks = self._rxns["tank"]

    @property
    def species(self) -> Dict[str, Species]:
        """Dictionary view onto only species"""
        return self._species

    @property
    def constants(self) -> Dict[str, Constant]:
        """Dictionary view onto only constants"""
        return self._const

    @property
    def parameters(self) -> Dict[str, Parameter]:
        """Dictionary view onto only parameters"""
        return self._param

    @property
    def terms(self) -> Dict[str, Term]:
        """Dictionary view onto only named terms"""
        return self._term

    @property
    def pipe_reactions(self) -> Dict[str, Reaction]:
        """Dictionary view onto pipe reactions"""
        return self._pipes

    @property
    def tank_reactions(self) -> Dict[str, Reaction]:
        """Dictionary view onto tank reactions"""
        return self._tanks

    def add_variable(self, variable: MsxVariable) -> None:
        """Add a variable object to the registry.

        The appropriate group is determined by querying the object's
        var_type attribute.

        Parameters
        ----------
        variable
            Variable to add.

        Raises
        ------
        TypeError
            If `variable` is not an MsxVariable
        KeyExistsError
            If `variable` has a name that is already used in the registry
        """
        if not isinstance(variable, (Species, Constant, Parameter, Term, MathFunction, HydraulicVariable)):
            raise TypeError("Expected AVariable object")
        if variable.name in self:
            raise KeyExistsError("Variable name {} already exists in model".format(variable.name))
        variable._vars = self
        self._vars.add_item_to_group(variable.var_type.name.lower(), variable.name, variable)

    def add_reaction(self, reaction: Reaction) -> None:
        """Add a reaction to the model

        Parameters
        ----------
        reaction : Reaction
            Water quality reaction definition

        Raises
        ------
        TypeError
            If `reaction` is not a Reaction
        KeyError
            If the `species_name` in the `reaction` does not exist in the model
        """
        if not isinstance(reaction, Reaction):
            raise TypeError("Expected a Reaction object")
        if reaction.species_name not in self:
            raise KeyError("Species {} does not exist in the model".format(reaction.species_name))
        self._rxns[reaction.reaction_type.name.lower()][reaction.species_name] = reaction

    def variables(self) -> Generator[tuple, None, None]:
        """Generator looping through all variables"""
        for k, v in self._vars.items():
            if v.var_type.name.lower() not in ['reserved']:
                yield k, v

    def reactions(self) -> Generator[tuple, None, None]:
        """Generator looping through all reactions"""
        for k2, v in self._rxns.items():
            for k1, v1 in v.items():
                yield k1, v1

    def to_dict(self) -> dict:
        """Dictionary representation of the MsxModel."""
        return dict(
            species=[v.to_dict() for v in self._species.values()],
            constants=[v.to_dict() for v in self._const.values()],
            parameters=[v.to_dict() for v in self._param.values()],
            terms=[v.to_dict() for v in self._term.values()],
            pipe_reactions=[v.to_dict() for v in self.pipe_reactions.values()],
            tank_reactions=[v.to_dict() for v in self.tank_reactions.values()],
        )


class MsxNetworkData(NetworkDataBase):
    """Network-specific values associated with a multi-species water 
    quality model

    Data is copied from dictionaries passed in, so once created, the
    dictionaries passed are not connected to this object.

    Parameters
    ----------
    patterns : dict, optional
        Patterns to use for sources
    sources : dict, optional
        Sources defined for the model
    initial_quality : dict, optional
        Initial values for different species at different nodes, links, and
        the global value
    parameter_values : dict, optional
        Parameter values for different pipes and tanks

    Notes
    -----
    ``patterns``
        Dictionary keyed by pattern name (str) with values being the
        multipliers (list of float)
    ``sources``
        Dictionary keyed by species name (str) with values being
        dictionaries keyed by junction name (str) with values being the
        dictionary of settings for the source
    ``initial_quality``
        Dictionary keyed by species name (str) with values being either an
        :class:`~wntr.msx.elements.InitialQuality` object or the
        appropriate dictionary representation thereof.
    ``parameter_values``
        Dictionary keyed by parameter name (str) with values being either
        a :class:`~wntr.msx.elements.ParameterValues` object or the
        appropriate dictionary representation thereof.
    """

    def __init__(self, patterns: Dict[str, List[float]] = None,
                 sources: Dict[str, Dict[str, dict]] = None,
                 initial_quality: Dict[str, Union[dict, InitialQuality]] = None,
                 parameter_values: Dict[str, Union[dict, ParameterValues]] = None) -> None:
        if sources is None:
            sources = dict()
        if initial_quality is None:
            initial_quality = dict()
        if patterns is None:
            patterns = dict()
        if parameter_values is None:
            parameter_values = dict()
        self._source_dict = dict()
        self._pattern_dict = dict()
        self._initial_quality_dict: Dict[str, InitialQuality] = dict()
        self._parameter_value_dict: Dict[str, ParameterValues] = dict()

        self._source_dict = sources.copy()
        self._pattern_dict = patterns.copy()
        for k, v in initial_quality.items():
            self._initial_quality_dict[k] = InitialQuality(**v)
        for k, v in parameter_values.items():
            self._parameter_value_dict[k] = ParameterValues(**v)

    @property
    def sources(self):
        """Dictionary of sources, keyed by species name"""
        return self._source_dict

    @property
    def initial_quality(self) -> Dict[str, InitialQuality]:
        """Dictionary of initial quality values, keyed by species name"""
        return self._initial_quality_dict

    @property
    def patterns(self):
        """Dictionary of patterns, specific for the water quality model, keyed
        by pattern name.

        .. note:: the WaterNetworkModel cannot see these patterns, so names can
           be reused, so be careful. Likewise, this model cannot see the
           WaterNetworkModel patterns, so this could be a source of some
           confusion.
        """
        return self._pattern_dict

    @property
    def parameter_values(self) -> Dict[str, ParameterValues]:
        """Dictionary of parameter values, keyed by parameter name"""
        return self._parameter_value_dict

    def add_pattern(self, name: str, multipliers: List[float]):
        """Add a water quality model specific pattern.

        Arguments
        ---------
        name : str
            Pattern name
        multipliers : list of float
            Pattern multipliers
        """
        self._pattern_dict[name] = multipliers

    def init_new_species(self, species: Species):
        """(Re)set the initial quality values for a species

        Arguments
        ---------
        species : Species
            Species to (re)initialized.

        Returns
        -------
        InitialQuality
            New initial quality values
        """
        self._initial_quality_dict[str(species)] = InitialQuality()
        if isinstance(species, Species):
            species._vals = self._initial_quality_dict[str(species)]
        return self._initial_quality_dict[str(species)]

    def remove_species(self, species: Union[Species, str]):
        """Remove a species from the network specific model

        Arguments
        ---------
        species : Species or str
            Species to be removed from the network data
        """
        if isinstance(species, Species):
            species._vals = None
        try:
            self._initial_quality_dict.__delitem__(str(species))
        except KeyError:
            pass

    def init_new_parameter(self, param: Parameter):
        """(Re)initialize parameter values for a parameter

        Arguments
        ---------
        param : Parameter
            Parameter to be (re)initialized with network data

        Returns
        -------
        ParameterValues
            New network data for the specific parameter
        """
        self._parameter_value_dict[str(param)] = ParameterValues()
        if isinstance(param, Parameter):
            param._vals = self._parameter_value_dict[str(param)]
        return self._parameter_value_dict[str(param)]

    def remove_parameter(self, param: Union[Parameter, str]):
        """Remove values associated with a specific parameter

        Ignores non-parameters.

        Arguments
        ---------
        param : Parameter or str
            Parameter or parameter name to be removed from the network data
        """
        if isinstance(param, Parameter):
            param._vals = None
        try:
            self._parameter_value_dict.__delitem__(str(param))
        except KeyError:
            pass

    def to_dict(self) -> dict:
        ret = dict(initial_quality=dict(), parameter_values=dict(), sources=dict(), patterns=dict())
        for k, v in self._initial_quality_dict.items():
            ret["initial_quality"][k] = v.to_dict()
        for k, v in self._parameter_value_dict.items():
            ret["parameter_values"][k] = v.to_dict()
        ret["sources"] = self._source_dict.copy()
        ret["patterns"] = self._pattern_dict.copy()
        return ret


class MsxModel(QualityModelBase):
    """Multi-species water quality model

    Arguments
    ---------
    msx_file_name : str, optional
        MSX file to to load into the MsxModel object, by default None
    """

    def __init__(self, msx_file_name=None) -> None:
        super().__init__(msx_file_name)
        self._references: List[Union[str, Dict[str, str]]] = list()
        self._options: MsxSolverOptions = MsxSolverOptions()
        self._rxn_system: MsxReactionSystem = MsxReactionSystem()
        self._net_data: MsxNetworkData = MsxNetworkData()
        self._wn = None

        for v in HYDRAULIC_VARIABLES:
            self._rxn_system.add_variable(HydraulicVariable(**v))
        for k, v in EXPR_FUNCTIONS.items():
            self._rxn_system.add_variable(MathFunction(name=k.lower(), func=v))
            self._rxn_system.add_variable(MathFunction(name=k.capitalize(), func=v))
            self._rxn_system.add_variable(MathFunction(name=k.upper(), func=v))

        if msx_file_name is not None:
            from wntr.epanet.msx.io import MsxFile

            MsxFile.read(msx_file_name, self)

    def __repr__(self) -> str:
        ret = "{}(".format(self.__class__.__name__)
        if self.name:
            ret = ret + "name={}".format(repr(self.name))
        elif self.title:
            ret = ret + "title={}".format(repr(self.title))
        elif self._orig_file:
            ret = ret + "{}".format(repr(self._orig_file))
        ret = ret + ")"
        return ret

    @property
    def references(self) -> List[Union[str, Dict[str, str]]]:
        """List of strings or mappings that provide references for this model

        .. note::
            This property is a list, and should be modified using
            append/insert/remove. Members of the list should be json
            serializable (i.e., strings or dicts of strings).
        """
        return self._references

    @property
    def reaction_system(self) -> MsxReactionSystem:
        """Reaction variables defined for this model"""
        return self._rxn_system

    @property
    def network_data(self) -> MsxNetworkData:
        """Network-specific values added to this model"""
        return self._net_data

    @property
    def options(self) -> MsxSolverOptions:
        """MSX model options"""
        return self._options

    @property
    def species_name_list(self) -> List[str]:
        """Get a list of species names"""
        return list(self.reaction_system.species.keys())

    @property
    def constant_name_list(self) -> List[str]:
        """Get a list of coefficient names"""
        return list(self.reaction_system.constants.keys())

    @property
    def parameter_name_list(self) -> List[str]:
        """Get a list of coefficient names"""
        return list(self.reaction_system.parameters.keys())

    @property
    def term_name_list(self) -> List[str]:
        """Get a list of function (MSX 'terms') names"""
        return list(self.reaction_system.terms.keys())

    @options.setter
    def options(self, value: Union[dict, MsxSolverOptions]):
        if isinstance(value, dict):
            self._options = MsxSolverOptions.factory(value)
        elif not isinstance(value, MsxSolverOptions):
            raise TypeError("Expected a MsxSolverOptions object, got {}".format(type(value)))
        else:
            self._options = value

    def add_species(
        self,
        name: str,
        species_type: SpeciesType,
        units: str,
        atol: float = None,
        rtol: float = None,
        note: NoteType = None,
        diffusivity: float = None,
    ) -> Species:
        """Add a species to the model

        Arguments
        ---------
        name : str
            Species name
        species_type : SpeciesType
            Type of species, either BULK or WALL
        units : str
            Mass units for  this species
        atol : float, optional unless rtol is not None
            Absolute solver tolerance for this species, by default None
        rtol : float, optional unless atol is not None
            Relative solver tolerance for this species, by default None
        note : NoteType, optional keyword
            Supplementary information regarding this variable, by default None
            (see also :class:`~wntr.epanet.util.ENcomment`)
        diffusivity : float, optional
            Diffusivity of this species in water

        Raises
        ------
        KeyExistsError
            If a variable with this name already exists
        ValueError
            If `atol` or `rtol` ≤ 0

        Returns
        -------
        Species
            New species
        """
        if name in self._rxn_system:
            raise KeyExistsError("Variable named {} already exists in model as type {}".format(name, self._rxn_system._vars.get_groupname(name)))
        species_type = SpeciesType.get(species_type, allow_none=False)
        iq = self.network_data.init_new_species(name)
        new = Species(
            name=name,
            species_type=species_type,
            units=units,
            atol=atol,
            rtol=rtol,
            note=note,
            _vars=self._rxn_system,
            _vals=iq,
            diffusivity=diffusivity,
        )
        self.reaction_system.add_variable(new)
        return new

    def remove_species(self, variable_or_name):
        """Remove a species from the model

        Removes from both the reaction_system and the network_data.

        Parameters
        ----------
        variable_or_name : Species or str
            Species (or name of the species) to be removed

        Raises
        ------
        KeyError
            If `variable_or_name` is not a species in the model
        """
        name = str(variable_or_name)
        if name not in self.reaction_system.species:
            raise KeyError('The specified variable is not a registered species in the reaction system')
        self.network_data.remove_species(name)
        self.reaction_system.__delitem__(name)

    def add_constant(self, name: str, value: float, units: str = None, note: NoteType = None) -> Constant:
        """Add a constant coefficient to the model

        Arguments
        ---------
        name : str
            Name of the coefficient
        value : float
            Constant value of the coefficient
        units : str, optional
            Units for this coefficient, by default None
        note : NoteType, optional
            Supplementary information regarding this variable, by default None

        Raises
        ------
        KeyExistsError
            Variable with this name already exists

        Returns
        -------
        Constant
            New constant coefficient
        """
        if name in self._rxn_system:
            raise KeyExistsError("Variable named {} already exists in model as type {}".format(name, self._rxn_system._vars.get_groupname(name)))
        new = Constant(name=name, value=value, units=units, note=note, _vars=self._rxn_system)
        self.reaction_system.add_variable(new)
        return new

    def remove_constant(self, variable_or_name):
        """Remove a constant coefficient from the model

        Parameters
        ----------
        variable_or_name : Constant or str
            Constant (or name of the constant) to be removed

        Raises
        ------
        KeyError
            If `variable_or_name` is not a constant coefficient in the model
        """
        name = str(variable_or_name)
        if name not in self.reaction_system.constants:
            raise KeyError('The specified variable is not a registered constant in the reaction system')
        self.reaction_system.__delitem__(name)

    def add_parameter(self, name: str, global_value: float, units: str = None, note: NoteType = None) -> Parameter:
        """Add a parameterized coefficient to the model

        Arguments
        ---------
        name : str
            Name of the parameter
        global_value : float
            Global value of the coefficient (can be overridden for specific
            pipes/tanks)
        units : str, optional
            Units for the coefficient, by default None
        note : NoteType, optional keyword
            Supplementary information regarding this variable, by default None
            (see also :class:`~wntr.epanet.util.ENcomment`).

        Raises
        ------
        KeyExistsError
            If a variable with this name already exists

        Returns
        -------
        Parameter
            New parameterized coefficient
        """
        if name in self._rxn_system:
            raise KeyExistsError("Variable named {} already exists in model as type {}".format(name, self._rxn_system._vars.get_groupname(name)))
        pv = self.network_data.init_new_parameter(name)
        new = Parameter(name=name, global_value=global_value, units=units, note=note, _vars=self._rxn_system, _vals=pv)
        self.reaction_system.add_variable(new)
        return new

    def remove_parameter(self, variable_or_name):
        """Remove a parameterized coefficient from the model

        Parameters
        ----------
        variable_or_name : Parameter or str
            Parameter (or name of the parameter) to be removed

        Raises
        ------
        KeyError
            If `variable_or_name` is not a parameter in the model
        """
        name = str(variable_or_name)
        if name not in self.reaction_system.parameters:
            raise KeyError('The specified variable is not a registered parameter in the reaction system')
        self.network_data.remove_parameter(name)
        self.reaction_system.__delitem__(name)

    def add_term(self, name: str, expression: str, note: NoteType = None) -> Term:
        """Add a named expression (term) to the model

        Parameters
        ----------
        name : str
            Name of the functional term to be added
        expression : str
            Expression that the term defines
        note : NoteType, optional keyword
            Supplementary information regarding this variable, by default None
            (see also :class:`~wntr.epanet.util.ENcomment`)

        Raises
        ------
        KeyExistsError
            if a variable with this name already exists

        Returns
        -------
        Term
            New term
        """
        if name in self._rxn_system:
            raise KeyError("Variable named {} already exists in model as type {}".format(name, self._rxn_system._vars.get_groupname(name)))
        new = Term(name=name, expression=expression, note=note, _vars=self._rxn_system)
        self.reaction_system.add_variable(new)
        return new

    def remove_term(self, variable_or_name):
        """Remove a named expression (term) from the model

        Parameters
        ----------
        variable_or_name : Term or str
            Term (or name of the term) to be deleted

        Raises
        ------
        KeyError
            If `variable_or_name` is not a term in the model
        """
        name = str(variable_or_name)
        if name not in self.reaction_system.terms:
            raise KeyError('The specified variable is not a registered term in the reaction system')
        self.reaction_system.__delitem__(name)

    def add_reaction(self, species_name: Union[Species, str], reaction_type: ReactionType, expression_type: ExpressionType, expression: str, note: NoteType = None) -> ReactionBase:
        """Add a reaction to a species in the model

        Note that all species need to have both a pipe and tank reaction
        defined unless all species are bulk species and the tank reactions are
        identical to the pipe reactions. However, it is not recommended that
        users take this approach.

        Once added, access the reactions from the species' object.

        Arguments
        ---------
        species_name : Species or str
            Species (or name of species) the reaction is being defined for
        reaction_type: ReactionType
            Reaction type (location), from {PIPE, TANK}
        expression_type : ExpressionType
            Expression type (left-hand-side) of the equation, from {RATE,
            EQUIL, FORMULA}
        expression : str
            Expression defining the reaction
        note : NoteType, optional keyword
            Supplementary information regarding this reaction, by default None
            (see also :class:`~wntr.epanet.util.ENcomment`)

        Raises
        ------
        TypeError
            If a variable that is not species is passed

        Returns
        -------
        MsxReactionSystem
            New reaction object
        """
        species_name = str(species_name)
        species = self.reaction_system.species[species_name]
        if species.var_type is not VariableType.SPECIES:
            raise TypeError("Variable {} is not a Species, is a {}".format(species.name, species.var_type))
        reaction_type = ReactionType.get(reaction_type, allow_none=False)
        expression_type = ExpressionType.get(expression_type, allow_none=False)
        new = Reaction(
            reaction_type=reaction_type,
            expression_type=expression_type,
            species_name=species_name,
            expression=expression,
            note=note,
        )
        self.reaction_system.add_reaction(new)
        return new

    def remove_reaction(self, species_name: str, reaction_type: ReactionType) -> None:
        """Remove a reaction at a specified location from a species

        Parameters
        ----------
        species : Species or str
            Species (or name of the species) of the reaction to remove
        reaction_type : ReactionType
            Reaction type (location) of the reaction to remove
        """
        reaction_type = ReactionType.get(reaction_type, allow_none=False)
        species_name = str(species_name)
        del self.reaction_system._rxns[reaction_type.name.lower()][species_name]
    
    def add_source(self, source_type: str, species: Union[Species, str], node: str, strength: Union[int, float], pattern: str = None, note: NoteType = None):
        """Add a new source of a specific msx species to the msx model.
        
        Arguments
        ---------
        source_type : str
            Source type, options = CONCEN, MASS, FLOWPACED, or SETPOINT
        species : str or Species
            Species object or name
        node : str or Node
            Juction, Tank or Reservoir object or name
        strength : float or int
            The strength of the species at the source
        pattern : str or Pattern
            Pattern name or object
        note : str or NoteType
            Note 
        """
        if str(species) not in self.network_data._source_dict:
            self.network_data._source_dict[str(species)] = {}

        d = {}
        d["strength"] = strength
        d["source_type"] = source_type
        d["pattern"] = pattern
        d["note"] = note
        self.network_data._source_dict[str(species)][str(node)] = d
        
    def remove_source(self, species: Union[Species, str], node: str):
        """Delete a source node from the msx object.
        
        Arguments
        ---------
        species : str or Species
            Species object or name
        """
        if str(species) not in self.network_data._source_dict:
            return
        if str(node) not in self.network_data._source_dict[str(species)]:
            return
        del self.network_data._source_dict[str(species)][str(node)]

    def remove_sources_for_species(self, species: Union[Species, str]):
        """Clear all sources for a specific species from the msx object.
        
        Arguments
        ---------
        species : str or Species
            Species object or name        
        """
        if str(species) in self.network_data._source_dict:
            del self.network_data._source_dict[str(species)]

    def get_sources_for_species(self, species: Union[Species, str]):
        """Get all sources for a specific species.
        
        Arguments
        ---------
        species : str or Species
            Species object or name

        Returns
        -------
        dict[str, float]
            mapping of node name to species strength
        """
        if str(species) not in self.network_data._source_dict:
            self.network_data._source_dict[str(species)] = {}
        return self.network_data._source_dict[str(species)]


    def to_dict(self) -> dict:
        """Dictionary representation of the MsxModel"""
        from wntr import __version__

        return {
            "version": "wntr-{}".format(__version__),
            "name": self.name,
            "title": self.title,
            "description": self.description if self.description is None or "\n" not in self.description else self.description.splitlines(),
            "references": self.references.copy(),
            "reaction_system": self.reaction_system.to_dict(),
            "network_data": self.network_data.to_dict(),
            "options": self.options.to_dict(),
        }

    @classmethod
    def from_dict(cls, data) -> "MsxModel":
        """Create a new multi-species water quality model from a dictionary
        
        Parameters
        ----------
        data : dict
            Model data
        """
        new = cls()
        new.name = data.get("name", None)
        new.title = data.get("title", None)
        new.description = data.get("description", None)
        if isinstance(new.description, (list, tuple)):
            desc = "\n".join(new.description)
            new.description = desc
        new.references.extend(data.get("references", list()))

        rxn_sys = data.get("reaction_system", dict())
        for var in rxn_sys.get("species", list()):
            new.add_species(**var)
        for var in rxn_sys.get("constants", list()):
            new.add_constant(**var)
        for var in rxn_sys.get("parameters", list()):
            new.add_parameter(**var)
        for var in rxn_sys.get("terms", list()):
            new.add_term(**var)
        for rxn in rxn_sys.get("pipe_reactions", list()):
            rxn["reaction_type"] = "pipe"
            new.add_reaction(**rxn)
        for rxn in rxn_sys.get("tank_reactions", list()):
            rxn["reaction_type"] = "tank"
            new.add_reaction(**rxn)

        new._net_data = MsxNetworkData(**data.get("network_data", dict()))
        for species in new.reaction_system.species:
            if species not in new.network_data.initial_quality:
                new.network_data.init_new_species(species)
        for param in new.reaction_system.parameters:
            if param not in new.network_data.parameter_values:
                new.network_data.init_new_parameter(param)

        opts = data.get("options", None)
        if opts:
            new.options = opts

        return new



# coding: utf-8
"""
A set of utility classes that is similar to the 'registry' objects in the wntr.network
class, but more general, and therefore usable for other extensions, such as multi-species
water quality models.
"""

from collections.abc import MutableMapping
from typing import Any, Dict, Hashable, ItemsView, Iterator, KeysView, Set, ValuesView


class WrongGroupError(KeyError):
    """The key exists but is in a different disjoint group"""
    pass


class KeyExistsError(KeyError):
    """The name already exists in the model"""
    pass


class DisjointMapping(MutableMapping):
    """A mapping with keys that are also divided into disjoint groups of keys.

    The main purpose of this utility class is to perform implicit name collision checking
    while also allowing both the groups and the main dictionary to be used as dictionaries
    -- i.e., using `__*item__` methods and `mydict[key]` methods.
    """

    __data: Dict[Hashable, Hashable] = None
    __key_groupnames: Dict[Hashable, str] = None
    __groups: Dict[str, "DisjointMappingGroup"] = None
    __usage: Dict[Hashable, Set[Any]] = None

    def __init__(self, *args, **kwargs):
        self.__data: Dict[Hashable, Any] = dict(*args, **kwargs)
        self.__key_groupnames: Dict[Hashable, str] = dict()
        self.__groups: Dict[str, "DisjointMappingGroup"] = dict()
        self.__usage: Dict[Hashable, Set[Any]] = dict()
        for k, v in self.__data.items():
            self.__key_groupnames[k] = None
            self.__usage[k] = set()

    def add_disjoint_group(self, name, cls = None):
        if name in self.__groups.keys():
            raise KeyError("Disjoint group already exists within registry")
        if cls is None:
            new = DisjointMappingGroup(name, self)
        elif issubclass(cls, DisjointMappingGroup):
            new = cls(name, self)
        else:
            raise TypeError('cls must be a subclass of DisjointMappingGroup, got {}'.format(cls))
        self.__groups.__setitem__(name, new)
        return new

    def get_disjoint_group(self, name: str):
        return self.__groups[name]

    def get_groupname(self, _key: Hashable):
        return self.__key_groupnames[_key]

    def add_item_to_group(self, groupname, key, value):
        current = self.__key_groupnames.get(key, None)
        if current is not None and groupname != current:
            raise WrongGroupError("The key '{}' is already used in a different group '{}'".format(key, groupname))
        if groupname is not None:
            group = self.__groups[groupname]
            group._data.__setitem__(key, value)
        self.__key_groupnames[key] = groupname
        return self.__data.__setitem__(key, value)

    def move_item_to_group(self, new_group_name, key):
        value = self.__data[key]
        current = self.get_groupname(key)
        if new_group_name is not None:
            new_group = self.__groups[new_group_name]
            new_group._data[key] = value
        if current is not None:
            old_group = self.__groups[current]
            old_group._data.__delitem__(key)
        self.__key_groupnames[key] = new_group_name

    def remove_item_from_group(self, groupname, key):
        current = self.__key_groupnames.get(key, None)
        if groupname != current:
            raise WrongGroupError("The key '{}' is in a different group '{}'".format(key, groupname))
        if groupname is not None:
            self.__groups[groupname]._data.__delitem__(key)

    def __getitem__(self, __key: Any) -> Any:
        return self.__data.__getitem__(__key)

    def __setitem__(self, __key: Any, __value: Any) -> None:
        current = self.__key_groupnames.get(__key, None)
        if current is not None:
            self.__groups[current]._data[__key] = __value
        return self.__data.__setitem__(__key, __value)

    def __delitem__(self, __key: Any) -> None:
        current = self.__key_groupnames.get(__key, None)
        if current is not None:
            self.__groups[current]._data.__delitem__(__key)
        return self.__data.__delitem__(__key)

    def __contains__(self, __key: object) -> bool:
        return self.__data.__contains__(__key)

    def __iter__(self) -> Iterator:
        return self.__data.__iter__()

    def __len__(self) -> int:
        return self.__data.__len__()

    def keys(self) -> KeysView:
        return self.__data.keys()

    def items(self) -> ItemsView:
        return self.__data.items()

    def values(self) -> ValuesView:
        return self.__data.values()

    def clear(self) -> None:
        raise RuntimeError("You cannot clear this")

    def popitem(self) -> tuple:
        raise RuntimeError("You cannot pop this")


class DisjointMappingGroup(MutableMapping):
    """A dictionary that checks a namespace for existing entries.

    To create a new instance, pass a set to act as a namespace. If the namespace does not
    exist, a new namespace will be instantiated. If it does exist, then a new, disjoint
    dictionary will be created that checks the namespace keys before it will allow a new
    item to be added to the dictionary. An item can only belong to one of the disjoint dictionaries
    associated with the namespace.

    Examples
    --------
    Assume there is a namespace `nodes` that has two distinct subsets of objects, `tanks`
    and `reservoirs`. A name for a tank cannot also be used for a reservoir, and a node
    cannot be both a `tank` and a `reservoir`. A DisjointNamespaceDict allows two separate
    dictionaries to be kept, one for each subtype, but the keys within the two dictionaries
    will be ensured to not overlap.

    Parameters
    ----------
    __keyspace : set
        the name of the namespace for consistency checking
    args, kwargs : Any
        regular arguments and keyword arguments passed to the underlying dictionary

    """

    __name: str = None
    __keyspace: "DisjointMapping" = None
    _data: dict = None

    def __new__(cls, name: str, _keyspace: "DisjointMapping"):
        if name is None:
            raise TypeError("A name must be specified")
        if _keyspace is None:
            raise TypeError("A registry must be specified")
        newobj = super().__new__(cls)
        return newobj

    def __init__(self, name: str, _keyspace: "DisjointMapping"):
        if name is None:
            raise TypeError("A name must be specified")
        if _keyspace is None:
            raise TypeError("A registry must be specified")
        self.__name: str = name
        self.__keyspace: DisjointMapping = _keyspace
        self._data = dict()

    def __getitem__(self, __key: Any) -> Any:
        return self._data[__key]

    def __setitem__(self, __key: Any, __value: Any) -> None:
        return self.__keyspace.add_item_to_group(self.__name, __key, __value)

    def __delitem__(self, __key: Any) -> None:
        return self.__keyspace.remove_item_from_group(self.__name, __key)

    def __contains__(self, __key: object) -> bool:
        return self._data.__contains__(__key)

    def __iter__(self) -> Iterator:
        return self._data.__iter__()

    def __len__(self) -> int:
        return self._data.__len__()

    def keys(self) -> KeysView:
        return self._data.keys()

    def items(self) -> ItemsView:
        return self._data.items()

    def values(self) -> ValuesView:
        return self._data.values()

    def clear(self) -> None:
        raise RuntimeError("Cannot clear a group")

    def popitem(self) -> tuple:
        raise RuntimeError("Cannot pop from a group")

    def __repr__(self) -> str:
        return "{}(name={}, data={})".format(self.__class__.__name__, repr(self.__name), self._data)



from matplotlib.colors import LinearSegmentedColormap, ListedColormap
import numpy as np
import matplotlib.pylab as plt
import logging

logger = logging.getLogger(__name__)

def custom_colormap(N, colors=['blue','white','red'], name='custom'):
    """ 
    Create a custom colormap.  Default settings creates a colormap named 'custom'
    which transitions from blue to white to red.
    
    Parameters
    -----------
    N : int 
        Number of bins in the colormap.
        
    colors : list of colors (optional)
        Colors can be specified in any way understandable by 
        matplotlib.colors.ColorConverter.to_rgb().  
    
    name : str (optional)
        Name of the colormap
        
    Returns
    --------
    cmap : matplotlib.colors.LinearSegmentedColormap object
    """
    
    cmap = LinearSegmentedColormap.from_list(name=name, 
                                             colors = colors,
                                             N=N)
    return cmap

def random_colormap(N, colormap='jet', name='random', seed=None):
    """ 
    Create a random ordered colormap.  Default settings creates a colormap named 'random'
    using the jet colormap.
    
    Parameters
    -----------
    N : int 
        Number of bins in the colormap.
        
    colormap : str (optional)
        Name of matplotlib colormap
    
    name : str (optional)
        Name of the colormap
    
    seed : int or None
        Random seed
        
    Returns
    --------
    cmap : matplotlib.colors.ListedColormap object
    """
    if seed is not None:
        np.random.seed(seed)
    
    vals = np.arange(0,1,1/N) 
    np.random.shuffle(vals)
    cmap = plt.get_cmap(colormap)
    cmap_random = ListedColormap(cmap(vals), name=name)

    return cmap_random



"""Utilitiy for inheriting docstrings."""

import inspect


class DocInheritor(object):
    def __init__(self, methods):
        self.methods = methods

    def __call__(self, cls):
        mro = inspect.getmro(cls)
        if mro[0] is not cls:
            raise RuntimeError('Unexpected')
        if len(mro) < 2:
            raise ValueError('Cannot determine parent class for {0}'.format(cls))
        parent = mro[1]
        for meth in self.methods:
            if not hasattr(parent, meth):
                raise ValueError('Parent class {0} does not have method {1}'.format(parent, meth))
            setattr(getattr(cls, meth), '__doc__', getattr(parent, meth).__doc__)
        return cls


# Example:
#
# class Foo(object):
#     def __init__(self, x):
#         """
#         Parameters
#         ----------
#         x: float
#         """
#         self.x = x
#
#     def inc(self):
#         """
#         increment x
#         """
#         self.x += 1
#
#
# @DocInheritor({'inc'})
# class Bar(Foo):
#     def inc(self):
#         self.x += 2



"""
The wntr.morph.link module contains functions to split/break pipes.
"""
import logging
import copy

import wntr.network
from wntr.network.elements import Reservoir, Pipe
from wntr.network import WaterNetworkModel

logger = logging.getLogger(__name__)


def split_pipe(wn, pipe_name_to_split, new_pipe_name, new_junction_name,
               add_pipe_at_end=True, split_at_point=0.5, return_copy=True):
    """
    Split a pipe by adding a junction and one new pipe segment.
    
    This function splits the original pipe into two pipes by adding a new 
    junction and new pipe to the model.  
    The updated model retains the original length of the pipe section. 
    The split occurs at a user specified distance between the 
    original start and end nodes of the pipe (in that direction). 
    The new pipe can be added to either end of the original pipe. 
    
    * The new junction has a base demand of 0 and the default demand pattern.
      The elevation and coordinates of the new junction are based on a linear 
      interpolation between the end points of the original pipe.
    
    * The new pipe has the same diameter, roughness, minor loss, 
      and base status of the original pipe. 

    * Check valves are not added to the new
      pipe. Since the new pipe can be connected at either the start
      or the end of the original pipe, the user can control if the split occurs before
      or after a check valve. 
    
    * No controls are added to the new pipe; the original pipe keeps any controls. 
    
    Parameters
    ----------
    wn: wntr WaterNetworkModel
        Water network model
    pipe_name_to_split: string
        Name of the pipe to split.
    new_pipe_name: string
        Name of the new pipe to be added as the split part of the pipe.
    new_junction_name: string
        Name of the new junction to be added.
    add_pipe_at_end: bool, optional
        If True, add the new pipe between the new node and the original end node. 
        If False, add the new pipe between the original start node and the new node.
    split_at_point: float, optional
        Between 0 and 1, the position along the original pipe where the new 
        junction will be located.
    return_copy: bool, optional
        If True, modify and return a copy of the WaterNetworkModel object.
        If False, modify and return the original WaterNetworkModel object.
        
    Returns
    -------
    wntr WaterNetworkModel
        Water network model with split pipe
    """
    wn2 = _split_or_break_pipe(wn, pipe_name_to_split, new_pipe_name,
                               [new_junction_name],
                               add_pipe_at_end, split_at_point, 'SPLIT', return_copy)

    return wn2


def break_pipe(wn, pipe_name_to_split, new_pipe_name, new_junction_name_old_pipe,
               new_junction_name_new_pipe, add_pipe_at_end=True,
               split_at_point=0.5, return_copy=True):
    """
    Break a pipe by adding a two unconnected junctions and one new pipe segment.
    
    This function splits the original pipe into two disconnected pipes by 
    adding two new junctions and new pipe to the model.  
    **This provides a true broken pipe -- i.e., there is no longer flow 
    possible from one side of the break to the other. This is more likely to 
    introduce non-convergable hydraulics than a simple split_pipe with a leak 
    added.**
    The updated model retains the original length of the pipe section. 
    The split occurs at a user specified distance between the 
    original start and end nodes of the pipe (in that direction). 
    The new pipe can be added to either end of the original pipe. 
    
    * The new junction has a base demand of 0 and the default demand pattern.
      The elevation and coordinates of the new junction are based on a linear 
      interpolation between the end points of the original pipe.
    
    * The new pipe has the same diameter, roughness, minor loss, 
      and base status of the original pipe. 

    * Check valves are not added to the new
      pipe. Since the new pipe can be connected at either the start
      or the end of the original pipe, the user can control if the split occurs before
      or after a check valve. 
    
    * No controls are added to the new pipe; the original pipe keeps any controls. 
    
    Parameters
    ----------
    wn: wntr WaterNetworkModel
        Water network model
    pipe_name_to_split: string
        Name of the pipe to split.
    new_pipe_name: string
        Name of the new pipe to be added as the split part of the pipe.
    new_junction_name_old_pipe: string
        Name of the new junction to be added to the original pipe
    new_junction_name_new_pipe: string
        Name of the new junction to be added to the new pipe
    add_pipe_at_end: bool, optional
        If True, add the new pipe at after the new junction. If False, add the 
        new pipe before the new junction
    split_at_point: float, optional
        Relative position (value between 0 and 1) along the original pipe 
        where the new junction will be located.
    return_copy: bool, optional
        If True, modify and return a copy of the WaterNetworkModel object.
        If False, modify and return the original WaterNetworkModel object.
        
    Returns
    -------
    wntr WaterNetworkModel
        Water network model with pipe break
    """
    wn2 = _split_or_break_pipe(wn, pipe_name_to_split, new_pipe_name,
                               [new_junction_name_old_pipe, new_junction_name_new_pipe],
                               add_pipe_at_end, split_at_point, 'BREAK', return_copy)

    return wn2


def _split_or_break_pipe(wn, pipe_name_to_split, new_pipe_name,
                         new_junction_names, add_pipe_at_end, split_at_point,
                         flag, return_copy):
    if return_copy:  # Get a copy of the WaterNetworkModel
        wn2 = copy.deepcopy(wn)
    else:
        wn2 = wn

    pipe = wn2.get_link(pipe_name_to_split)

    # Do sanity checks
    if not isinstance(pipe, Pipe):
        raise ValueError('You can only split pipes.')
    if split_at_point < 0 or split_at_point > 1:
        raise ValueError('split_at_point must be between 0 and 1')
    node_list = [node_name for node_name, node in wn2.nodes()]
    link_list = [link_name for link_name, link in wn2.links()]
    for new_junction_name in new_junction_names:
        if new_junction_name in node_list:
            raise RuntimeError('The junction name you provided is already \
                               being used for another node.')
    if new_pipe_name in link_list:
        raise RuntimeError('The new link name you provided is already being \
                           used for another link.')

    # Get start and end node info
    start_node = pipe.start_node
    end_node = pipe.end_node

    # calculate the new elevation
    if isinstance(start_node, Reservoir):
        junction_elevation = end_node.elevation
    elif isinstance(end_node, Reservoir):
        junction_elevation = start_node.elevation
    else:
        e0 = start_node.elevation
        de = end_node.elevation - e0
        junction_elevation = e0 + de * split_at_point

    # calculate the new coordinates and where each vertice belongs
    first_vertices = []
    last_vertices = []
    if pipe.vertices:
        # Extract pipe vertices including the start and end node coordinate
        pipe_vertices = [pipe.start_node.coordinates,
                         *pipe.vertices,
                         pipe.end_node.coordinates]
        segments = []
        subtotal = 0
        for i in range(len(pipe_vertices) - 1):
            start_pos = pipe_vertices[i]
            end_pos = pipe_vertices[i + 1]
            segment_length = (sum([(a - b) ** 2 for (a, b) in zip(start_pos, end_pos)])
                              ** 0.5)
            segments.append({'start_pos': start_pos,
                             'end_pos': end_pos,
                             'length': segment_length,
                             'subtotal': subtotal})

            subtotal += segment_length

        last_segment = segments[-1]
        length = last_segment['subtotal'] + last_segment['length']
        split_length = length * split_at_point

        # Loop over segments and assign vertices (start_pos) to the first or 
        # second pipe. Skip the first segment (its start_pos is not a vertice)
        for segment in segments:
            if segment['start_pos'] == pipe.start_node.coordinates:
                pass
            elif segment['subtotal'] < split_length:
                first_vertices.append(segment['start_pos'])
            else:
                last_vertices.append(segment['start_pos'])

        # Identify the segment that cross the split point and compute the new
        # junction coordinates
        for segment in segments:
            if segment['subtotal'] + segment['length'] >= split_length > segment['subtotal']:
                split_at = ((split_length - segment['subtotal'])
                            / segment['length'])
                x0 = segment['start_pos'][0]
                dx = segment['end_pos'][0] - x0
                y0 = segment['start_pos'][1]
                dy = segment['end_pos'][1] - y0
                junction_coordinates = (x0 + dx * split_at,
                                        y0 + dy * split_at)
    # calculate the new coordinates without vertices            
    else:
        x0 = pipe.start_node.coordinates[0]
        dx = pipe.end_node.coordinates[0] - x0
        y0 = pipe.start_node.coordinates[1]
        dy = pipe.end_node.coordinates[1] - y0
        junction_coordinates = (x0 + dx * split_at_point,
                                y0 + dy * split_at_point)

    # add the new junction
    for new_junction_name in new_junction_names:
        wn2.add_junction(new_junction_name, base_demand=0.0,
                         demand_pattern=None, elevation=junction_elevation,
                         coordinates=junction_coordinates)

    original_length = pipe.length

    if flag == 'BREAK':
        j0 = new_junction_names[0]
        j1 = new_junction_names[1]
    elif flag == 'SPLIT':
        j0 = new_junction_names[0]
        j1 = new_junction_names[0]

    if add_pipe_at_end:
        pipe.end_node = wn2.get_node(j0)
        # add new pipe and change original length
        wn2.add_pipe(new_pipe_name, j1, end_node.name,
                     original_length * (1 - split_at_point), pipe.diameter,
                     pipe.roughness, pipe.minor_loss, pipe.status, pipe.check_valve)
        pipe.length = original_length * split_at_point
        pipe.vertices = first_vertices
        new_pipe = wn2.get_link(new_pipe_name)
        new_pipe.vertices = last_vertices
    else:  # add pipe at start
        pipe.start_node = wn2.get_node(j0)
        # add new pipe and change original length
        wn2.add_pipe(new_pipe_name, start_node.name, j1,
                     original_length * split_at_point, pipe.diameter,
                     pipe.roughness, pipe.minor_loss, pipe.status, pipe.check_valve)
        pipe.length = original_length * (1 - split_at_point)
        pipe.vertices = last_vertices
        new_pipe = wn2.get_link(new_pipe_name)
        new_pipe.vertices = first_vertices

    if pipe.check_valve:
        logger.warn('You are splitting a pipe with a check valve. The new \
                    pipe will not have a check valve.')

    return wn2


def reverse_link(wn: WaterNetworkModel, link_name: str, return_copy=True) -> WaterNetworkModel:
    """
    Reverse a link to switch between the start and end nodes
    The function can reverse any link (pipe, pump or valve)
    Vertices order will be reversed to maintain the link layout

    Parameters
    ----------
    wn:             WaterNetworkModel
                    The network object where link should be reversed.
    link_name:      string
                    The name of the link to revers.
    return_copy:    bool, optional
                    If True, modify and return a copy of the WaterNetworkModel object.
                    If False, modify and return the original WaterNetworkModel object.

    Returns
    -------
    WaterNetworkModel
    A network object after link was reversed
    """
    if return_copy:  # Get a copy of the WaterNetworkModel
        wn2 = copy.deepcopy(wn)
    else:
        wn2 = wn

    link = wn2.get_link(link_name)
    start_node = link.start_node
    end_node = link.end_node
    vertices = link.vertices

    link.start_node = end_node
    link.end_node = start_node
    link.vertices = vertices[::-1]

    return wn2



# coding: utf-8
"""
The wntr.epanet.msx.exceptions module contains Exceptions for EPANET-MSX
IO operations.
"""

from typing import List

from ..exceptions import EpanetException

MSX_ERROR_CODES = {
    # MSX syntax errors
    401: "too many characters",
    402: "too few input items",
    403: "invalid keyword: '%s'",
    404: "invalid numeric value: '%s'",
    405: "reference to undefined object: '%s'",
    406: "illegal use of reserved name: '%s'",
    407: "name already used by another object: '%s'",
    408: "species already assigned an expression: '%s'",
    409: "illegal math expression",
    410: "option no longer supported",
    411: "term '%s' contains a cyclic reference",
    # MSX runtime errors
    501: "insufficient memory available",
    502: "no EPANET data file supplied",
    503: "could not open MSX input file %s",
    504: "could not open hydraulic results file",
    505: "could not read hydraulic results file",
    506: "could not read MSX input file %s",
    507: "too few pipe reaction expressions",
    508: "too few tank reaction expressions",
    509: "could not open differential equation solver",
    510: "could not open algebraic equation solver",
    511: "could not open binary results file",
    512: "read/write error on binary results file",
    513: "could not integrate reaction rate expressions",
    514: "could not solve reaction equilibrium expressions",
    515: "reference made to an unknown type of object %s",
    516: "reference made to an illegal object index %s",
    517: "reference made to an undefined object ID %s",
    518: "invalid property values were specified",
    519: "an MSX project was not opened",
    520: "an MSX project is already opened",
    522: "could not compile chemistry functions",
    523: "could not load functions from compiled chemistry file",
    524: "illegal math operation",
}
"""A dictionary of the error codes and their meanings from the EPANET-MSX toolkit.

:meta hide-value:
"""


class EpanetMsxException(EpanetException):
    def __init__(self, code: int, *args: List[object], line_num=None,
                 line=None) -> None:
        """Exception class for EPANET-MSX Toolkit and IO exceptions

        Parameters
        ----------
        code : int or str or MSXErrors
            EPANET-MSX error code (int) or a string mapping to the MSXErrors
            enum members
        args : additional non-keyword arguments, optional
            If there is a string-format within the error code's text, these
            will be used to replace the format, otherwise they will be output
            at the end of the Exception message.
        line_num : int, optional
            Line number, if reading an INP file, by default None
        line : str, optional
            Contents of the line, by default None
        """
        if not code or int(code) < 400:
            return super().__init__(code, *args, line_num=line_num, line=line)
        msg = MSX_ERROR_CODES.get(code, "unknown MSX error number {}".format(code))
        if args is not None:
            args = [*args]
        if r"%" in msg and len(args) > 0:
            msg = msg % args.pop(0)
        if len(args) > 0:
            msg = msg + " " + str(args)
        if line_num:
            msg = msg + ", at line {}".format(line_num)
        if line:
            msg = msg + ":\n   " + str(line)
        msg = "(Error {}) ".format(code) + msg
        super(Exception, self).__init__(msg)


class MSXSyntaxError(EpanetMsxException, SyntaxError):
    def __init__(self, code, *args, line_num=None, line=None) -> None:
        """MSX-specific error that is due to a syntax error in an msx-file.

        Parameters
        ----------
        code : int or str or MSXErrors
            EPANET-MSX error code (int) or a string mapping to the MSXErrors
            enum members
        args : additional non-keyword arguments, optional
            If there is a string-format within the error code's text, these
            will be used to replace the format, otherwise they will be output
            at the end of the Exception message.
        line_num : int, optional
            Line number, if reading an INP file, by default None
        line : str, optional
            Contents of the line, by default None
        """
        super().__init__(code, *args, line_num=line_num, line=line)


class MSXKeyError(EpanetMsxException, KeyError):
    def __init__(self, code, name, *args, line_num=None, line=None) -> None:
        """MSX-specific error that is due to a missing or unacceptably named
        variable/speces/etc.

        Parameters
        ----------
        code : int or str or MSXErrors
            EPANET-MSX error code (int) or a string mapping to the MSXErrors
            enum members
        name : str
            Key/name/id that is missing
        args : additional non-keyword arguments, optional
            If there is a string-format within the error code's text, these
            will be used to replace the format, otherwise they will be output
            at the end of the Exception message.
        line_num : int, optional
            Line number, if reading an INP file, by default None
        line : str, optional
            Contents of the line, by default None
        """
        super().__init__(code, name, *args, line_num=line_num, line=line)


class MSXValueError(EpanetMsxException, ValueError):
    def __init__(self, code, value, *args, line_num=None, line=None) -> None:
        """MSX-specific error that is related to an invalid value.

        Parameters
        ----------
        code : int or str or MSXErrors
            EPANET-MSX error code (int) or a string mapping to the MSXErrors
            enum members
        value : Any
            Value that is invalid
        args : additional non-keyword arguments, optional
            If there is a string-format within the error code's text, these
            will be used to replace the format, otherwise they will be output
            at the end of the Exception message.
        line_num : int, optional
            Line number, if reading an INP file, by default None
        line : str, optional
            Contents of the line, by default None
        """

        super().__init__(code, value, *args, line_num=line_num, line=line)



from setuptools import setup, find_packages
from setuptools.extension import Extension
import os
import re
import sys

use_swig = False
build = True

if '--no-build' in sys.argv:
    build = False
    sys.argv.remove('--no-build')

extension_modules = list()
if build:
    import numpy

    try:
        numpy_include = numpy.get_include()
    except AttributeError:
        numpy_include = numpy.get_numpy_include()

    # inplace extension module
    project_dir = './'  # os.path.dirname(os.path.abspath(__file__))
    src_files = os.path.join(project_dir, 'wntr', 'sim', 'aml')
    evaluator_cxx = os.path.join(src_files, 'evaluator.cpp')
    evaluator_wrap_cxx = os.path.join(src_files, 'evaluator_wrap.cpp')
    evaluator_i = os.path.join(src_files, 'evaluator.i')
    network_isolation_dir = os.path.join(project_dir, 'wntr', 'sim', 'network_isolation')
    network_isolation_cxx = os.path.join(network_isolation_dir, 'network_isolation.cpp')
    network_isolation_i = os.path.join(network_isolation_dir, 'network_isolation.i')
    network_isolation_wrap_cxx = os.path.join(network_isolation_dir, 'network_isolation_wrap.cpp')

    if use_swig:
        aml_core_ext = Extension("wntr.sim.aml._evaluator",
                                 sources=[evaluator_i, evaluator_cxx],
                                 language="c++",
                                 extra_compile_args=[],
                                 include_dirs=[numpy_include, src_files],
                                 library_dirs=[],
                                 libraries=[],
                                 swig_opts = ['-c++', '-builtin'])
        network_isolation_ext = Extension("wntr.sim.network_isolation._network_isolation",
                                          sources=[network_isolation_i, network_isolation_cxx],
                                          language="c++",
                                          include_dirs=[numpy_include, network_isolation_dir],
                                          extra_compile_args=[],
                                          swig_opts=['-c++', '-builtin'])
    else:
        aml_core_ext = Extension("wntr.sim.aml._evaluator",
                                 sources=[evaluator_cxx, evaluator_wrap_cxx],
                                 language="c++",
                                 extra_compile_args=[],
                                 include_dirs=[numpy_include, src_files],
                                 library_dirs=[],
                                 libraries=[])
        network_isolation_ext = Extension("wntr.sim.network_isolation._network_isolation",
                                          sources=[network_isolation_cxx, network_isolation_wrap_cxx],
                                          language="c++",
                                          include_dirs=[numpy_include, network_isolation_dir],
                                          extra_compile_args=[])


    extension_modules.append(aml_core_ext)
    extension_modules.append(network_isolation_ext)

DISTNAME = 'wntr'
PACKAGES = find_packages()
EXTENSIONS = extension_modules
DESCRIPTION = 'Water Network Tool for Resilience'
AUTHOR = 'WNTR Developers'
MAINTAINER_EMAIL = 'kaklise@sandia.gov'
LICENSE = 'Revised BSD'
URL = 'https://github.com/USEPA/WNTR'
DEPENDENCIES = ['numpy>=2.2.6 ', 'scipy', 'networkx', 'pandas', 'matplotlib', 'setuptools']

# use README file as the long description
file_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(file_dir, 'README.md'), encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()

# get version from __init__.py
with open(os.path.join(file_dir, 'wntr', '__init__.py')) as f:
    version_file = f.read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        VERSION = version_match.group(1)
    else:
        raise RuntimeError("Unable to find version string.")

print(extension_modules)

setup(name=DISTNAME,
      version=VERSION,
      packages=PACKAGES,
      ext_modules=EXTENSIONS,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      author=AUTHOR,
      maintainer_email=MAINTAINER_EMAIL,
      license=LICENSE,
      url=URL,
      zip_safe=False,
      install_requires=DEPENDENCIES,
      scripts=[],
      include_package_data=True)



import unittest
from os.path import abspath, dirname, join

testdir = dirname(abspath(str(__file__)))
test_datadir = join(testdir, "networks_for_testing")
ex_datadir = join(testdir, "..", "..", "examples", "networks")

# class TestNetworkTimeWarnings(unittest.TestCase):
#
#    @classmethod
#    def setUpClass(self):
#        import wntr
#        self.wntr = wntr
#
#    @classmethod
#    def tearDownClass(self):
#        pass
#
#    def test_pattern_start_time(self):
#        inp_file = join(test_datadir, 'net_test_8.inp')
#        wn = self.wntr.network.WaterNetworkModel()
#        parser = self.wntr.network.ParseWaterNetwork()
#
#        flag = False
#        with warnings.catch_warnings(record=True) as w:
#            warnings.simplefilter("always")
#            parser.read_inp_file(wn, inp_file)
#        for message in w:
#            if str(message.message) == 'Currently, only the EpanetSimulator supports a non-zero patern start time.':
#                flag = True
#        self.assertEqual(flag, True)
#
#    #def test_report_time_step(self):
#    #    inp_file = join(test_datadir, 'net_test_9.inp')
#    #    wn = self.wntr.network.WaterNetworkModel()
#    #    parser = self.wntr.network.ParseWaterNetwork()
#    #
#    #    flag = False
#    #    with warnings.catch_warnings(record=True) as w:
#    #        warnings.simplefilter("always")
#    #        parser.read_inp_file(wn, inp_file)
#    #    for message in w:
#    #        if str(message.message) == 'Currently, only a the EpanetSimulator supports a report timestep that is not equal to the hydraulic timestep.':
#    #            flag = True
#    #    self.assertEqual(flag, True)
#
#    def test_report_start_time(self):
#        inp_file = join(test_datadir, 'net_test_10.inp')
#        wn = self.wntr.network.WaterNetworkModel()
#        parser = self.wntr.network.ParseWaterNetwork()
#
#        flag = False
#        with warnings.catch_warnings(record=True) as w:
#            warnings.simplefilter("always")
#            parser.read_inp_file(wn, inp_file)
#        for message in w:
#            if str(message.message) == 'Currently, only the EpanetSimulator supports a non-zero report start time.':
#                flag = True
#        self.assertEqual(flag, True)
#
#    def test_start_clocktime(self):
#        inp_file = join(test_datadir, 'net_test_11.inp')
#        wn = self.wntr.network.WaterNetworkModel()
#        parser = self.wntr.network.ParseWaterNetwork()
#
#        flag = False
#        with warnings.catch_warnings(record=True) as w:
#            warnings.simplefilter("always")
#            parser.read_inp_file(wn, inp_file)
#        for message in w:
#            if str(message.message) == 'Currently, only the EpanetSimulator supports a start clocktime other than 12 am.':
#                flag = True
#        self.assertEqual(flag, True)


class TestNetworkTimeBehavior(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        import wntr

        self.wntr = wntr

        inp_file = join(test_datadir, "times.inp")
        wn = self.wntr.network.WaterNetworkModel(inp_file)
        sim = self.wntr.sim.WNTRSimulator(wn)
        self.results = sim.run_sim()

    @classmethod
    def tearDownClass(self):
        pass

    def test_duration(self):
        results = self.results
        self.assertEqual(len(results.node["head"].index), 26)
        self.assertEqual(results.node["head"].index[25], 24 * 3600 + 3 * 3600 + 5 * 60)

    def test_report_timestep(self):
        results = self.results
        self.assertEqual(
            (results.node["head"].index[1] - results.node["head"].index[0]),
            1 * 3600 + 5 * 60,
        )

    def test_pattern_timestep(self):
        results = self.results
        for t in results.node["demand"].index:
            self.assertEqual(results.node["demand"].at[t, "junction1"], 1.0)
            total_seconds = t
            if (total_seconds / 3900.0) % 8 == 0.0 or (
                (total_seconds / 3900.0) - 1
            ) % 8 == 0.0:
                self.assertEqual(results.node["demand"].at[t, "junction2"], 0.5)
            elif (total_seconds / 3900.0) % 8 == 2.0 or (
                (total_seconds / 3900.0) - 1
            ) % 8 == 2.0:
                self.assertEqual(results.node["demand"].at[t, "junction2"], 1.0)
            elif (total_seconds / 3900.0) % 8 == 4.0 or (
                (total_seconds / 3900.0) - 1
            ) % 8 == 4.0:
                self.assertEqual(results.node["demand"].at[t, "junction2"], 1.5)
            elif (total_seconds / 3900.0) % 8 == 6.0 or (
                (total_seconds / 3900.0) - 1
            ) % 8 == 6.0:
                self.assertEqual(results.node["demand"].at[t, "junction2"], 1.0)


if __name__ == "__main__":
    unittest.main()
